# stdlib
import json
from typing import Dict
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# syft absolute
from syft import serialize
from syft.core.common.message import ImmediateSyftMessageWithReply
from syft.core.common.message import ImmediateSyftMessageWithoutReply
from syft.core.common.serde.deserialize import _deserialize
from syft.core.common.serde.serializable import bind_protobuf
from syft.core.common.uid import UID
from syft.core.io.address import Address
from syft.proto.grid.messages.group_messages_pb2 import (
    CreateGroupMessage as CreateGroupMessage_PB,
)
from syft.proto.grid.messages.group_messages_pb2 import (
    CreateGroupResponse as CreateGroupResponse_PB,
)
from syft.proto.grid.messages.group_messages_pb2 import (
    DeleteGroupMessage as DeleteGroupMessage_PB,
)
from syft.proto.grid.messages.group_messages_pb2 import (
    DeleteGroupResponse as DeleteGroupResponse_PB,
)
from syft.proto.grid.messages.group_messages_pb2 import (
    GetGroupMessage as GetGroupMessage_PB,
)
from syft.proto.grid.messages.group_messages_pb2 import (
    GetGroupResponse as GetGroupResponse_PB,
)
from syft.proto.grid.messages.group_messages_pb2 import (
    GetGroupsMessage as GetGroupsMessage_PB,
)
from syft.proto.grid.messages.group_messages_pb2 import (
    GetGroupsResponse as GetGroupsResponse_PB,
)
from syft.proto.grid.messages.group_messages_pb2 import (
    UpdateGroupMessage as UpdateGroupMessage_PB,
)
from syft.proto.grid.messages.group_messages_pb2 import (
    UpdateGroupResponse as UpdateGroupResponse_PB,
)


@bind_protobuf
@final
class CreateGroupMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> CreateGroupMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: CreateGroupMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return CreateGroupMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: CreateGroupMessage_PB,
    ) -> "CreateGroupMessage":
        """Creates a CreateGroupMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: CreateGroupMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return CreateGroupMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return CreateGroupMessage_PB


@bind_protobuf
@final
class CreateGroupResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> CreateGroupResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return CreateGroupResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: CreateGroupResponse_PB,
    ) -> "CreateGroupResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return CreateGroupResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return CreateGroupResponse_PB


@bind_protobuf
@final
class GetGroupMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> GetGroupMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetGroupMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetGroupMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetGroupMessage_PB,
    ) -> "GetGroupMessage":
        """Creates a GetGroupMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetGroupMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetGroupMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetGroupMessage_PB


@bind_protobuf
@final
class GetGroupResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> GetGroupResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetGroupResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: GetGroupResponse_PB,
    ) -> "GetGroupResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetGroupResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetGroupResponse_PB


@bind_protobuf
@final
class GetGroupsMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> GetGroupsMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetGroupsMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetGroupsMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetGroupsMessage_PB,
    ) -> "GetGroupsMessage":
        """Creates a GetGroupsMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetGroupsMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetGroupsMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetGroupsMessage_PB


@bind_protobuf
@final
class GetGroupsResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> GetGroupsResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetGroupsResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: GetGroupsResponse_PB,
    ) -> "GetGroupsResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetGroupsResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetGroupsResponse_PB


@bind_protobuf
@final
class UpdateGroupMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> UpdateGroupMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: UpdateGroupMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return UpdateGroupMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: UpdateGroupMessage_PB,
    ) -> "UpdateGroupMessage":
        """Creates a UpdateGroupMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: UpdateGroupMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return UpdateGroupMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return UpdateGroupMessage_PB


@bind_protobuf
@final
class UpdateGroupResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> UpdateGroupResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return UpdateGroupResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: UpdateGroupResponse_PB,
    ) -> "UpdateGroupResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return UpdateGroupResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return UpdateGroupResponse_PB


@bind_protobuf
@final
class DeleteGroupMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> DeleteGroupMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: DeleteGroupMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return DeleteGroupMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: DeleteGroupMessage_PB,
    ) -> "DeleteGroupMessage":
        """Creates a DeleteGroupMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: DeleteGroupMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return DeleteGroupMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return DeleteGroupMessage_PB


@bind_protobuf
@final
class DeleteGroupResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> DeleteGroupResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return DeleteGroupResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: DeleteGroupResponse_PB,
    ) -> "DeleteGroupResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return DeleteGroupResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return DeleteGroupResponse_PB



# syft absolute
import syft as sy


def test_path_cache() -> None:
    short_fqn_list = "syft.lib.python.List"
    long_fqn_list = "syft.lib.python.list.List"

    list_ref1 = sy.lib_ast.query(short_fqn_list, obj_type=sy.lib.python.List)
    list_ref2 = sy.lib_ast.query(long_fqn_list, obj_type=sy.lib.python.List)
    list_ref2 = sy.lib_ast.query(long_fqn_list)

    assert list_ref1.object_ref == sy.lib.python.List
    assert list_ref1.name == "List"
    assert list_ref1.path_and_name == short_fqn_list

    assert list_ref2.object_ref == sy.lib.python.List
    assert list_ref2.name == "List"
    assert list_ref2.path_and_name == short_fqn_list

    assert list_ref1 == list_ref2



# third party
import pytest

# syft absolute
import syft as sy


@pytest.mark.vendor(lib="pydp")
def test_pydp(root_client: sy.VirtualMachineClient) -> None:
    sy.load("pydp")
    x_ptr = root_client.pydp.algorithms.laplacian.BoundedMean(1, 1, 50)

    input_data = [1, 88, 100, 5, 40, 30, 29, 56, 88, 23, 5, 1] * 100
    list_ptr = root_client.python.List(input_data)

    res_ptr = x_ptr.quick_result(list_ptr)
    res = res_ptr.get()

    assert 22 < res < 35


@pytest.mark.vendor(lib="pydp")
def test_pydp_functions(root_client: sy.VirtualMachineClient) -> None:
    sy.load("pydp")
    x_ptr = root_client.pydp.algorithms.laplacian.BoundedMean(1, 1, 50)

    input_data = [1, 88, 100, 5, 40, 30, 29, 56, 88, 23, 5, 1] * 100
    list_ptr = root_client.python.List(input_data)
    x_ptr.add_entries(list_ptr)
    res_ptr = x_ptr.result(0.7)
    assert isinstance(res_ptr.get(), float)

    assert round(x_ptr.privacy_budget_left().get(), 2) == 0.3



# third party
import pytest
import torch as th

# syft absolute
from syft.core.common.message import SyftMessage
from syft.core.node.domain import Domain
from syft.core.node.domain.service import RequestStatus


@pytest.mark.asyncio
async def test_domain_creation() -> None:
    Domain(name="test domain")


@pytest.mark.asyncio
def test_domain_serde() -> None:

    domain_1 = Domain(name="domain 1")
    domain_1_client = domain_1.get_client()

    tensor = th.tensor([1, 2, 3])
    _ = tensor.send(domain_1_client)


@pytest.mark.asyncio
def test_domain_request_pending() -> None:
    domain_1 = Domain(name="remote domain")
    tensor = th.tensor([1, 2, 3])

    domain_1_client = domain_1.get_root_client()
    data_ptr_domain_1 = tensor.send(domain_1_client)

    domain_2 = Domain(name="my domain")

    data_ptr_domain_1.request(
        reason="I'd lke to see this pointer",
    )

    requested_object = data_ptr_domain_1.id_at_location

    # make request
    message_request_id = domain_1_client.requests.get_request_id_from_object_id(
        object_id=requested_object
    )

    # check status
    response = data_ptr_domain_1.check_access(
        node=domain_2, request_id=message_request_id
    )

    assert RequestStatus.Pending == response


@pytest.mark.slow
@pytest.mark.asyncio
def test_domain_request_denied() -> None:
    domain_1 = Domain(name="remote domain")
    tensor = th.tensor([1, 2, 3])

    domain_1_client = domain_1.get_root_client()
    data_ptr_domain_1 = tensor.send(domain_1_client)

    domain_2 = Domain(name="my domain")

    data_ptr_domain_1.request(reason="I'd lke to see this pointer")

    requested_object = data_ptr_domain_1.id_at_location

    # make request
    message_request_id = domain_1_client.requests.get_request_id_from_object_id(
        object_id=requested_object
    )

    # domain 1 client rejects request
    domain_1.requests[0].owner_client_if_available = domain_1_client
    domain_1.requests[0].deny()

    # check status
    response = data_ptr_domain_1.check_access(
        node=domain_2, request_id=message_request_id
    )

    assert RequestStatus.Rejected == response


@pytest.mark.asyncio
def test_domain_request_accepted() -> None:
    domain_1 = Domain(name="remote domain")
    tensor = th.tensor([1, 2, 3])

    domain_1_client = domain_1.get_root_client()
    data_ptr_domain_1 = tensor.send(domain_1_client)

    domain_2 = Domain(name="my domain")

    data_ptr_domain_1.request(reason="I'd lke to see this pointer")

    requested_object = data_ptr_domain_1.id_at_location

    message_request_id = domain_1_client.requests.get_request_id_from_object_id(
        object_id=requested_object
    )

    domain_1.requests[0].owner_client_if_available = domain_1_client
    domain_1.requests[0].accept()

    response = data_ptr_domain_1.check_access(
        node=domain_2, request_id=message_request_id
    )

    assert RequestStatus.Accepted == response


@pytest.mark.asyncio
def test_domain_is_for_me_exception() -> None:
    domain_1 = Domain(name="remote domain")

    with pytest.raises(Exception):
        msg = SyftMessage()
        domain_1.message_is_for_me(msg)



# stdlib
import argparse
import os
import re
import sys
import time

# third party
import numpy as np
import torch
import torch.onnx
from torch.optim import Adam
from torch.utils.data import DataLoader
from torchvision import datasets
from torchvision import transforms
from transformer_net import TransformerNet
import utils
from vgg import Vgg16


def check_paths(args):
    try:
        if not os.path.exists(args.save_model_dir):
            os.makedirs(args.save_model_dir)
        if args.checkpoint_model_dir is not None and not (
            os.path.exists(args.checkpoint_model_dir)
        ):
            os.makedirs(args.checkpoint_model_dir)
    except OSError as e:
        print(e)
        sys.exit(1)


def train(args):
    device = torch.device("cuda" if args.cuda else "cpu")

    np.random.seed(args.seed)
    torch.manual_seed(args.seed)

    transform = transforms.Compose(
        [
            transforms.Resize(args.image_size),
            transforms.CenterCrop(args.image_size),
            transforms.ToTensor(),
            transforms.Lambda(lambda x: x.mul(255)),
        ]
    )
    train_dataset = datasets.ImageFolder(args.dataset, transform)
    train_loader = DataLoader(train_dataset, batch_size=args.batch_size)

    transformer = TransformerNet().to(device)
    optimizer = Adam(transformer.parameters(), args.lr)
    mse_loss = torch.nn.MSELoss()

    vgg = Vgg16(requires_grad=False).to(device)
    style_transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Lambda(lambda x: x.mul(255))]
    )
    style = utils.load_image(args.style_image, size=args.style_size)
    style = style_transform(style)
    style = style.repeat(args.batch_size, 1, 1, 1).to(device)

    features_style = vgg(utils.normalize_batch(style))
    gram_style = [utils.gram_matrix(y) for y in features_style]

    for e in range(args.epochs):
        transformer.train()
        agg_content_loss = 0.0
        agg_style_loss = 0.0
        count = 0
        for batch_id, (x, _) in enumerate(train_loader):
            n_batch = len(x)
            count += n_batch
            optimizer.zero_grad()

            x = x.to(device)
            y = transformer(x)

            y = utils.normalize_batch(y)
            x = utils.normalize_batch(x)

            features_y = vgg(y)
            features_x = vgg(x)

            content_loss = args.content_weight * mse_loss(
                features_y.relu2_2, features_x.relu2_2
            )

            style_loss = 0.0
            for ft_y, gm_s in zip(features_y, gram_style):
                gm_y = utils.gram_matrix(ft_y)
                style_loss += mse_loss(gm_y, gm_s[:n_batch, :, :])
            style_loss *= args.style_weight

            total_loss = content_loss + style_loss
            total_loss.backward()
            optimizer.step()

            agg_content_loss += content_loss.item()
            agg_style_loss += style_loss.item()

            if (batch_id + 1) % args.log_interval == 0:
                mesg = "{}\tEpoch {}:\t[{}/{}]\tcontent: {:.6f}\tstyle: {:.6f}\ttotal: {:.6f}".format(
                    time.ctime(),
                    e + 1,
                    count,
                    len(train_dataset),
                    agg_content_loss / (batch_id + 1),
                    agg_style_loss / (batch_id + 1),
                    (agg_content_loss + agg_style_loss) / (batch_id + 1),
                )
                print(mesg)

            if (
                args.checkpoint_model_dir is not None
                and (batch_id + 1) % args.checkpoint_interval == 0
            ):
                transformer.eval().cpu()
                ckpt_model_filename = (
                    "ckpt_epoch_" + str(e) + "_batch_id_" + str(batch_id + 1) + ".pth"
                )
                ckpt_model_path = os.path.join(
                    args.checkpoint_model_dir, ckpt_model_filename
                )
                torch.save(transformer.state_dict(), ckpt_model_path)
                transformer.to(device).train()

    # save model
    transformer.eval().cpu()
    save_model_filename = (
        "epoch_"
        + str(args.epochs)
        + "_"
        + str(time.ctime()).replace(" ", "_")
        + "_"
        + str(args.content_weight)
        + "_"
        + str(args.style_weight)
        + ".model"
    )
    save_model_path = os.path.join(args.save_model_dir, save_model_filename)
    torch.save(transformer.state_dict(), save_model_path)

    print("\nDone, trained model saved at", save_model_path)


def stylize(args):
    device = torch.device("cuda" if args.cuda else "cpu")

    content_image = utils.load_image(args.content_image, scale=args.content_scale)
    content_transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Lambda(lambda x: x.mul(255))]
    )
    content_image = content_transform(content_image)
    content_image = content_image.unsqueeze(0).to(device)

    if args.model.endswith(".onnx"):
        output = stylize_onnx_caffe2(content_image, args)
    else:
        with torch.no_grad():
            style_model = TransformerNet()
            state_dict = torch.load(args.model)
            # remove saved deprecated running_* keys in InstanceNorm from the checkpoint
            for k in list(state_dict.keys()):
                if re.search(r"in\d+\.running_(mean|var)$", k):
                    del state_dict[k]
            style_model.load_state_dict(state_dict)
            style_model.to(device)
            if args.export_onnx:
                assert args.export_onnx.endswith(
                    ".onnx"
                ), "Export model file should end with .onnx"
                output = torch.onnx._export(
                    style_model, content_image, args.export_onnx
                ).cpu()
            else:
                output = style_model(content_image).cpu()
    utils.save_image(args.output_image, output[0])


def stylize_onnx_caffe2(content_image, args):
    """
    Read ONNX model and run it using Caffe2
    """

    assert not args.export_onnx

    # third party
    import onnx
    import onnx_caffe2.backend

    model = onnx.load(args.model)

    prepared_backend = onnx_caffe2.backend.prepare(
        model, device="CUDA" if args.cuda else "CPU"
    )
    inp = {model.graph.input[0].name: content_image.numpy()}
    c2_out = prepared_backend.run(inp)[0]

    return torch.from_numpy(c2_out)


def main():
    main_arg_parser = argparse.ArgumentParser(
        description="parser for fast_neural_style"
    )
    subparsers = main_arg_parser.add_subparsers(title="subcommands", dest="subcommand")

    train_arg_parser = subparsers.add_parser(
        "train", help="parser for training arguments"
    )
    train_arg_parser.add_argument(
        "--epochs", type=int, default=2, help="number of training epochs; default: 2"
    )
    train_arg_parser.add_argument(
        "--batch_size",
        type=int,
        default=4,
        help="batch size for training; default: 4",
    )
    train_arg_parser.add_argument(
        "--dataset",
        type=str,
        required=True,
        help="path to training dataset, the path to a folder containing another folder with the training images",
    )
    train_arg_parser.add_argument(
        "--style_image",
        type=str,
        default="images/style_images/mosaic.jpg",
        help="path to style_image",
    )
    train_arg_parser.add_argument(
        "--save_model_dir",
        type=str,
        required=True,
        help="path to folder where trained model will be saved.",
    )
    train_arg_parser.add_argument(
        "--checkpoint_model_dir",
        type=str,
        default=None,
        help="path to folder where checkpoints of trained models will be saved",
    )
    train_arg_parser.add_argument(
        "--image_size",
        type=int,
        default=256,
        help="size of training images; default: 256 X 256",
    )
    train_arg_parser.add_argument(
        "--style_size",
        type=int,
        default=None,
        help="size of style_image; default: the original size of style image",
    )
    train_arg_parser.add_argument(
        "--cuda",
        type=int,
        required=True,
        help="set to 1 for running on GPU, 0 for CPU",
    )
    train_arg_parser.add_argument(
        "--seed", type=int, default=42, help="random seed for training"
    )
    train_arg_parser.add_argument(
        "--content_weight",
        type=float,
        default=1e5,
        help="weight for content_loss; default: 1e5",
    )
    train_arg_parser.add_argument(
        "--style_weight",
        type=float,
        default=1e10,
        help="weight for style_loss; default: 1e10",
    )
    train_arg_parser.add_argument(
        "--lr", type=float, default=1e-3, help="learning rate; default: 1e-3"
    )
    train_arg_parser.add_argument(
        "--log_interval",
        type=int,
        default=500,
        help="number of images, after which the training loss is logged; default: 500",
    )
    train_arg_parser.add_argument(
        "--checkpoint_interval",
        type=int,
        default=2000,
        help="number of batches, after which a checkpoint for the trained model will be created; default:2000",
    )

    eval_arg_parser = subparsers.add_parser(
        "eval", help="parser for evaluation/stylizing arguments"
    )
    eval_arg_parser.add_argument(
        "--content_image",
        type=str,
        required=True,
        help="path to content image to stylize",
    )
    eval_arg_parser.add_argument(
        "--content_scale",
        type=float,
        default=None,
        help="factor for scaling down the content image",
    )
    eval_arg_parser.add_argument(
        "--output_image",
        type=str,
        required=True,
        help="path for saving the output image",
    )
    eval_arg_parser.add_argument(
        "--model",
        type=str,
        required=True,
        help="saved model to be used for stylizing the image. If file ends in .pth - PyTorch path is used, if in .onnx - Caffe2 path",
    )
    eval_arg_parser.add_argument(
        "--cuda",
        type=int,
        required=True,
        help="set it to 1 for running on GPU, 0 for CPU",
    )
    eval_arg_parser.add_argument(
        "--export_onnx", type=str, help="export ONNX model to a given file"
    )

    args = main_arg_parser.parse_args()

    if args.subcommand is None:
        print("ERROR: specify either train or eval")
        sys.exit(1)
    if args.cuda and not torch.cuda.is_available():
        print("ERROR: cuda is not available, try running on CPU")
        sys.exit(1)

    if args.subcommand == "train":
        check_paths(args)
        train(args)
    else:
        stylize(args)


if __name__ == "__main__":
    main()



# third party
import petlib as pl

# syft relative
from ...core.common.serde.deserialize import _deserialize as deserialize
from ...core.common.serde.serialize import _serialize as serialize
from ...generate_wrapper import GenerateWrapper
from ...proto.lib.petlib.ecpt_pb2 import EcPt as EcPt_PB


def object2proto(obj: pl.ec.EcPt) -> EcPt_PB:
    return EcPt_PB(group=serialize(obj.group, to_proto=True), data=obj.export())


def proto2object(proto: EcPt_PB) -> pl.ec.EcPt:
    eg = deserialize(blob=proto.group, from_proto=True)
    return pl.ec.EcPt.from_binary(proto.data, eg)


GenerateWrapper(
    wrapped_type=pl.ec.EcPt,
    import_path="petlib.ec.EcPt",
    protobuf_scheme=EcPt_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# stdlib
from typing import Any
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# syft relative
from ...logger import traceback_and_raise
from ...proto.core.store.store_object_pb2 import StorableObject as StorableObject_PB
from ...util import get_fully_qualified_name
from ...util import index_syft_by_module_name
from ...util import key_emoji
from ..common.serde.deserialize import _deserialize
from ..common.serde.serializable import Serializable
from ..common.serde.serializable import bind_protobuf
from ..common.storeable_object import AbstractStorableObject
from ..common.uid import UID


@bind_protobuf
class StorableObject(AbstractStorableObject):
    """
    StorableObject is a wrapper over some Serializable objects, which we want to keep in an
    ObjectStore. The Serializable objects that we want to store have to be backed up in syft-proto
    in the StorableObject protobuffer, where you can find more details on how to add new types to be
    serialized.

    This object is frozen, you cannot change one in place.

    Arguments:
        id (UID): the id at which to store the data.
        data (Serializable): A serializable object.
        description (Optional[str]): An optional string that describes what you are storing. Useful
        when searching.
        tags (Optional[List[str]]): An optional list of strings that are tags used at search.
        TODO: add docs about read_permission and search_permission

    Attributes:
        id (UID): the id at which to store the data.
        data (Serializable): A serializable object.
        description (Optional[str]): An optional string that describes what you are storing. Useful
        when searching.
        tags (Optional[List[str]]): An optional list of strings that are tags used at search.

    """

    __slots__ = ["id", "_data", "_description", "_tags"]

    def __init__(
        self,
        id: UID,
        data: object,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        read_permissions: Optional[dict] = None,
        search_permissions: Optional[dict] = None,
    ):
        self.id = id
        self.data = data
        self._description: str = description if description else ""
        self._tags: List[str] = tags if tags else []

        # the dict key of "verify key" objects corresponding to people
        # the value is the original request_id to allow lookup later
        # who are allowed to call .get() and download this object.
        self.read_permissions = read_permissions if read_permissions else {}

        # the dict key of "verify key" objects corresponding to people
        # the value is the original request_id to allow lookup later
        # who are allowed to know that the tensor exists (via search or other means)
        self.search_permissions: dict = search_permissions if search_permissions else {}

    @property
    def object_type(self) -> str:
        return str(type(self.data))

    @property
    def object_qualname(self) -> str:
        return get_fully_qualified_name(self.data)

    # Why define data as a property?
    # For C type/class objects as data.
    # We need to use it's wrapper type very often inside StorableObject, so we set _data
    # attribute as it's wrapper object. But we still want to give a straight API to users,
    # so we return the initial C type object when user call obj.data.
    # For python class objects as data. data and _data are the same thing.
    @property  # type: ignore
    def data(self) -> Any:  # type: ignore
        if type(self._data).__name__.endswith("Wrapper"):
            return self._data.obj
        else:
            return self._data

    @data.setter
    def data(self, value: Any) -> Any:
        if hasattr(value, "_sy_serializable_wrapper_type"):
            self._data = value._sy_serializable_wrapper_type(value=value)
        else:
            self._data = value

    @property
    def tags(self) -> Optional[List[str]]:
        return self._tags

    @tags.setter
    def tags(self, value: Optional[List[str]]) -> None:
        self._tags = value if value else []

    @property
    def description(self) -> Optional[str]:
        return self._description

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self._description = description if description else ""

    def _object2proto(self) -> StorableObject_PB:
        proto = StorableObject_PB()

        # Step 1: Serialize the id to protobuf and copy into protobuf
        id = sy.serialize(self.id)
        proto.id.CopyFrom(id)

        # Step 2: Save the type of wrapper to use to deserialize
        proto.data_type = get_fully_qualified_name(obj=self._data)

        # Step 3: Serialize data to protobuf and pack into proto
        data = self._data._object2proto()

        proto.data.Pack(data)

        if hasattr(self, "description"):
            # Step 4: save the description into proto
            proto.description = self.description

        # QUESTION: Which one do we want, self.data.tags or self.tags or both???
        if hasattr(self, "tags"):
            # Step 5: save tags into proto if they exist
            if self.tags is not None:
                for tag in self.tags:
                    proto.tags.append(tag)

        # Step 6: save read permissions
        if len(self.read_permissions) > 0:
            permission_data = sy.lib.python.Dict()
            for k, v in self.read_permissions.items():
                permission_data[k] = v
            proto.read_permissions = sy.serialize(permission_data, to_bytes=True)

        # Step 7: save search permissions
        if len(self.search_permissions.keys()) > 0:
            permission_data = sy.lib.python.Dict()
            for k, v in self.search_permissions.items():
                permission_data[k] = v
            proto.search_permissions = sy.serialize(permission_data, to_bytes=True)

        return proto

    @staticmethod
    def _proto2object(proto: StorableObject_PB) -> Serializable:
        # Step 1: deserialize the ID
        id = _deserialize(blob=proto.id)

        if not isinstance(id, UID):
            traceback_and_raise(ValueError("TODO"))

        # Step 2: get the type of wrapper to use to deserialize
        data_type = index_syft_by_module_name(fully_qualified_name=proto.data_type)

        # Step 3: get the protobuf type we deserialize for .data
        schematic_type = data_type.get_protobuf_schema()  # type: ignore

        # Step 4: Deserialize data from protobuf
        data = None
        if callable(schematic_type):
            data = schematic_type()
            descriptor = getattr(schematic_type, "DESCRIPTOR", None)
            if descriptor is not None and proto.data.Is(descriptor):
                proto.data.Unpack(data)
            data = data_type._proto2object(proto=data)  # type: ignore

        # Step 5: get the description from proto
        description = proto.description if proto.description else ""

        # Step 6: get the tags from proto of they exist
        tags = list(proto.tags) if proto.tags else []

        # Step 7: get the read permissions
        read_permissions = None
        if proto.read_permissions is not None and len(proto.read_permissions) > 0:
            read_permissions = _deserialize(
                blob=proto.read_permissions, from_bytes=True
            )

        # Step 8: get the search permissions
        search_permissions = None
        if proto.search_permissions is not None and len(proto.search_permissions) > 0:
            search_permissions = _deserialize(
                blob=proto.search_permissions, from_bytes=True
            )

        result = StorableObject(
            id=id,
            data=data,
            description=description,
            tags=tags,
            read_permissions=read_permissions,
            search_permissions=search_permissions,
        )

        return result

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """
        return StorableObject_PB

    def __repr__(self) -> str:
        return (
            "<Storable: "
            + self.data.__repr__().replace("\n", "").replace("  ", " ")
            + ">"
        )

    @property
    def icon(self) -> str:
        return "🗂️"

    @property
    def pprint(self) -> str:
        output = f"{self.icon} ({self.class_name}) ("
        if hasattr(self.data, "pprint"):
            output += self.data.pprint
        elif self.data is not None:
            output += self.data.__repr__()
        else:
            output += "(Key Only)"
        if len(self._description) > 0:
            output += f" desc: {self.description}"
        if len(self._tags) > 0:
            output += f" tags: {self.tags}"
        if len(self.read_permissions.keys()) > 0:
            output += (
                " can_read: "
                + f"{[key_emoji(key=key) for key in self.read_permissions.keys()]}"
            )

        if len(self.search_permissions.keys()) > 0:
            output += (
                " can_search: "
                + f"{[key_emoji(key=key) for key in self.search_permissions.keys()]}"
            )

        output += ")"
        return output

    @property
    def class_name(self) -> str:
        return str(self.__class__.__name__)

    def clean_copy(self) -> "StorableObject":
        """
        This method return a copy of self, but clean up the search_permissions and
        read_permissions attributes.
        """
        return StorableObject(
            id=self.id, data=self.data, tags=self.tags, description=self.description
        )



# third party
from pytest import raises

# syft absolute
from syft.core.common.serde.deserialize import _deserialize


def test_fail_deserialize_no_format() -> None:
    with raises(TypeError):
        _deserialize(blob="to deserialize", from_proto=False)


def test_fail_deserialize_wrong_format() -> None:
    with raises(TypeError, match="You tried to deserialize an unsupported type."):
        _deserialize(blob="to deserialize")



# stdlib
import sys
from typing import Generic
from typing import Optional
from typing import Type
from typing import TypeVar

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.exceptions import BadSignatureError
from nacl.signing import SigningKey
from nacl.signing import VerifyKey

# syft relative
from ...core.common.object import ObjectWithID
from ...core.common.serde.serialize import _serialize as serialize
from ...core.common.uid import UID
from ...core.io.address import Address
from ...logger import debug
from ...logger import traceback_and_raise
from ...proto.core.auth.signed_message_pb2 import SignedMessage as SignedMessage_PB
from ...util import get_fully_qualified_name
from ...util import validate_type
from ..common.serde.deserialize import _deserialize
from ..common.serde.serializable import bind_protobuf

# this generic type for SignedMessage
SignedMessageT = TypeVar("SignedMessageT")


class AbstractMessage(ObjectWithID, Generic[SignedMessageT]):
    """"""

    # this should be overloaded by a subclass
    signed_type: Type[SignedMessageT]

    @property
    def class_name(self) -> str:
        return str(self.__class__.__name__)

    @property
    def icon(self) -> str:
        icon = "✉️ "
        if "signed" in self.class_name.lower():
            icon += "🔏"
        return icon

    @property
    def pprint(self) -> str:
        return f"{self.icon} ({self.class_name})"

    def post_init(self) -> None:
        init_reason = "Creating"
        if "signed" in self.class_name.lower():
            init_reason += " Signed"
        debug(f"> {init_reason} {self.pprint} {self.id.emoji()}")


class SyftMessage(AbstractMessage):
    """
    SyftMessages are an abstraction that represents information that is sent between a :class:`Client`
    and a :class:`Node`. In Syft's decentralized setup, we can easily see why SyftMessages are so important.
    This class cannot be used as-is: to get some useful objects, we need to derive from it. For instance,
    :class:`Action`s inherit from :class:`SyftMessage`.
    There are many types of SyftMessage which boil down to whether or not they are Sync or Async,
    and whether or not they expect a response.

    Attributes:
        address: the :class:`Address` to which the message needs to be delivered.
    """

    def __init__(self, address: Address, msg_id: Optional[UID] = None) -> None:
        self.address = address
        super().__init__(id=msg_id)
        self.post_init()

    def sign(self, signing_key: SigningKey) -> SignedMessageT:
        """
        It's important for all messages to be able to prove who they were sent from.
        This method endows every message with the ability for someone to "sign" (with a hash of the message)
        as the sender of the message so that someone else, at a later date, can verify the sender.

        Args:
            signing_key: The key to use to sign the SyftMessage.

        Returns:
            A :class:`SignedMessage`

        """
        debug(f"> Signing with {self.address.key_emoji(key=signing_key.verify_key)}")
        signed_message = signing_key.sign(serialize(self, to_bytes=True))

        # signed_type will be the final subclass callee's closest parent signed_type
        # for example ReprMessage -> ImmediateSyftMessageWithoutReply.signed_type
        # == SignedImmediateSyftMessageWithoutReply
        return self.signed_type(
            msg_id=self.id,
            address=self.address,
            obj_type=get_fully_qualified_name(obj=self),
            signature=signed_message.signature,
            verify_key=signing_key.verify_key,
            message=signed_message.message,
        )


@bind_protobuf
class SignedMessage(SyftMessage):
    """
    SignedMessages are :class:`SyftMessage`s that have been signed by someone.
    In addition to what has a :class:`SyftMessage`, they have a signature, a verify key
    and a :meth:`is_valid` property that is here to check that the message was really
    signed and sent by the verify key owner.

    Attributes:
        obj_type (string): the string representation of the type of the original message.
        signature (bytes): the signature of the message.
        verify_key (VerifyKey): the signer's public key with which the signature can be verified.
        serialized_message: the serialized original message.
    """

    obj_type: str
    signature: bytes
    verify_key: VerifyKey

    def __init__(
        self,
        address: Address,
        obj_type: str,
        signature: bytes,
        verify_key: VerifyKey,
        message: bytes,
        msg_id: Optional[UID] = None,
    ) -> None:
        super().__init__(msg_id=msg_id, address=address)
        self.obj_type = obj_type
        self.signature = signature
        self.verify_key = verify_key
        self.serialized_message = message
        self.cached_deseralized_message: Optional[SyftMessage] = None

    @property
    def message(self) -> "SyftMessage":
        if self.cached_deseralized_message is None:
            _syft_msg = validate_type(
                _deserialize(blob=self.serialized_message, from_bytes=True), SyftMessage
            )
            self.cached_deseralized_message = _syft_msg

        if self.cached_deseralized_message is None:
            traceback_and_raise(
                ValueError(
                    f"Can't deserialize message {self} with address " f"{self.address}"
                )
            )

        return self.cached_deseralized_message

    @property
    def is_valid(self) -> bool:
        try:
            _ = self.verify_key.verify(self.serialized_message, self.signature)
        except BadSignatureError:
            return False

        return True

    def _object2proto(self) -> SignedMessage_PB:
        debug(f"> {self.icon} -> Proto 🔢 {self.id}")

        # obj_type will be the final subclass callee for example ReprMessage
        return SignedMessage_PB(
            msg_id=serialize(self.id, to_proto=True),
            obj_type=self.obj_type,
            signature=bytes(self.signature),
            verify_key=bytes(self.verify_key),
            message=self.serialized_message,
        )

    @staticmethod
    def _proto2object(proto: SignedMessage_PB) -> SignedMessageT:
        # TODO: horrible temp hack, need to rethink address on SignedMessage
        sub_message = validate_type(
            _deserialize(blob=proto.message, from_bytes=True), SyftMessage
        )

        address = sub_message.address

        # proto.obj_type is final subclass callee for example ReprMessage
        # but we want the associated signed_type which is
        # ReprMessage -> ImmediateSyftMessageWithoutReply.signed_type
        # == SignedImmediateSyftMessageWithoutReply
        module_parts = proto.obj_type.split(".")
        klass = module_parts.pop()
        obj_type = getattr(sys.modules[".".join(module_parts)], klass)
        obj = obj_type.signed_type(
            msg_id=_deserialize(blob=proto.msg_id),
            address=address,
            obj_type=proto.obj_type,
            signature=proto.signature,
            verify_key=VerifyKey(proto.verify_key),
            message=proto.message,
        )

        icon = "🤷🏾‍♀️"
        if hasattr(obj, "icon"):
            icon = obj.icon
        debug(f"> {icon} <- 🔢 Proto")

        if type(obj) != obj_type.signed_type:
            traceback_and_raise(
                TypeError(
                    "Deserializing SignedMessage. "
                    + f"Expected type {obj_type.signed_type}. Got {type(obj)}"
                )
            )

        return obj

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return SignedMessage_PB


class SignedImmediateSyftMessageWithReply(SignedMessage):
    """"""


class SignedImmediateSyftMessageWithoutReply(SignedMessage):
    """"""

    # do stuff


class SignedEventualSyftMessageWithoutReply(SignedMessage):
    """"""


class ImmediateSyftMessage(SyftMessage):
    ""


class EventualSyftMessage(SyftMessage):
    ""


class SyftMessageWithReply(SyftMessage):
    ""


class SyftMessageWithoutReply(SyftMessage):
    ""


class ImmediateSyftMessageWithoutReply(ImmediateSyftMessage, SyftMessageWithoutReply):

    signed_type = SignedImmediateSyftMessageWithoutReply

    def __init__(self, address: Address, msg_id: Optional[UID] = None) -> None:
        super().__init__(address=address, msg_id=msg_id)


class EventualSyftMessageWithoutReply(EventualSyftMessage, SyftMessageWithoutReply):

    signed_type = SignedEventualSyftMessageWithoutReply

    def __init__(self, address: Address, msg_id: Optional[UID] = None) -> None:
        super().__init__(address=address, msg_id=msg_id)


class ImmediateSyftMessageWithReply(ImmediateSyftMessage, SyftMessageWithReply):

    signed_type = SignedImmediateSyftMessageWithReply

    def __init__(
        self, reply_to: Address, address: Address, msg_id: Optional[UID] = None
    ) -> None:
        super().__init__(address=address, msg_id=msg_id)
        self.reply_to = reply_to



# syft absolute
from syft.core.common.message import SignedEventualSyftMessageWithoutReply
from syft.core.common.message import SignedImmediateSyftMessageWithReply
from syft.core.common.message import SignedImmediateSyftMessageWithoutReply
from syft.core.io.location.specific import SpecificLocation
from syft.core.io.route import BroadcastRoute
from syft.core.io.route import Route
from syft.core.io.route import RouteSchema
from syft.core.io.route import SoloRoute
from syft.core.io.virtual import VirtualClientConnection
from syft.core.io.virtual import VirtualServerConnection
from syft.core.node.common.node import Node

# syft relative
from .utils_test import MockNode
from .utils_test import construct_dummy_message

# --------------------- INITIALIZATION ---------------------


def test_route_schema_init() -> None:
    """Test that RouteSchema will use the Location passed into its constructor"""
    destination = SpecificLocation()
    rschema = RouteSchema(destination)

    assert rschema.destination is not None
    assert rschema.destination._id == destination._id


def test_route_init() -> None:
    """Test that Route will use the RouteSchema passed into its constructor"""
    rschema = RouteSchema(SpecificLocation())
    route = Route(schema=rschema)

    assert route.schema is rschema  # Cannot use __eq__
    assert isinstance(route.stops, list)


def test_solo_route_init() -> None:
    """
    Test that SoloRoute will use the Location and
    ClientConnection/BidirectionalConnection passed into its constructor.
    """
    # Test SoloRoute with VirtualClientConnection (ClientConnection) in constructor
    destination = SpecificLocation()
    virtual_server = VirtualServerConnection(node=Node())
    virtual_client = VirtualClientConnection(server=virtual_server)
    h_solo = SoloRoute(destination=destination, connection=virtual_client)

    assert h_solo.schema.destination is destination
    assert h_solo.connection is virtual_client


def test_broadcast_route_init() -> None:
    """
    Test that BroadcastRoute will use the Location and
    ClientConnection passed into its constructor.
    """
    destination = SpecificLocation()
    virtual_server = VirtualServerConnection(node=Node())
    virtual_client = VirtualClientConnection(server=virtual_server)
    b_route = BroadcastRoute(destination=destination, connection=virtual_client)

    assert b_route.schema.destination is destination
    assert b_route.connection is virtual_client


# --------------------- CLASS & PROPERTY METHODS ---------------------


def test_route_icon_property_method() -> None:
    """Test Route.icon property method returns the correct icon."""
    route = Route(schema=RouteSchema(SpecificLocation()))
    assert route.icon == "🛣️ "


def test_route_pprint_property_method() -> None:
    """Test Route.pprint property method returns the correct icon."""
    route = Route(schema=RouteSchema(SpecificLocation()))
    assert route.pprint == "🛣️  (Route)"


def test_solo_route_send_immediate_msg_without_reply() -> None:
    """Test SoloRoute.send_immediate_msg_without_reply method works."""
    node = MockNode()
    destination = SpecificLocation()
    server = VirtualServerConnection(node=node)
    connection = VirtualClientConnection(server=server)

    h_solo = SoloRoute(destination=destination, connection=connection)
    msg = construct_dummy_message(SignedImmediateSyftMessageWithoutReply)

    assert h_solo.send_immediate_msg_without_reply(msg) is None


def test_solo_route_send_eventual_msg_without_reply() -> None:
    """Test SoloRoute.send_eventual_msg_without_reply method works."""
    node = MockNode()
    destination = SpecificLocation()
    server = VirtualServerConnection(node=node)
    connection = VirtualClientConnection(server=server)

    h_solo = SoloRoute(destination=destination, connection=connection)
    msg = construct_dummy_message(SignedEventualSyftMessageWithoutReply)

    assert not h_solo.send_eventual_msg_without_reply(msg)


def test_solo_route_send_immediate_msg_with_reply() -> None:
    """Test SoloRoute.send_immediate_msg_with_reply method works."""
    node = MockNode()
    destination = SpecificLocation()
    server = VirtualServerConnection(node=node)
    connection = VirtualClientConnection(server=server)

    h_solo = SoloRoute(destination=destination, connection=connection)
    msg = construct_dummy_message(SignedImmediateSyftMessageWithReply)
    ret = h_solo.send_immediate_msg_with_reply(msg)

    assert isinstance(
        ret,
        SignedImmediateSyftMessageWithoutReply,
    )



# -*- coding: utf-8 -*-
"""
CODING GUIDELINES:

Do NOT (without talking to trask):
- add another high level method for sending or receiving messages (like recv_eventual_msg_without_reply)
- add a service to the list of services below unless you're SURE all nodes will need it!
- serialize anything with pickle
"""

# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey

# syft relative
from ....lib import lib_ast
from ....logger import critical
from ....logger import debug
from ....logger import error
from ....logger import traceback_and_raise
from ....util import get_subclasses
from ...common.message import EventualSyftMessageWithoutReply
from ...common.message import ImmediateSyftMessageWithReply
from ...common.message import ImmediateSyftMessageWithoutReply
from ...common.message import SignedEventualSyftMessageWithoutReply
from ...common.message import SignedImmediateSyftMessageWithReply
from ...common.message import SignedImmediateSyftMessageWithoutReply
from ...common.message import SignedMessage
from ...common.message import SyftMessage
from ...common.uid import UID
from ...io.address import Address
from ...io.location import Location
from ...io.route import Route
from ...io.route import SoloRoute
from ...io.virtual import create_virtual_connection
from ...store import DiskObjectStore
from ...store import MemoryStore
from ..abstract.node import AbstractNode
from .action.exception_action import ExceptionMessage
from .action.exception_action import UnknownPrivateException
from .client import Client
from .metadata import Metadata
from .service.auth import AuthorizationException
from .service.child_node_lifecycle_service import ChildNodeLifecycleService
from .service.get_repr_service import GetReprService
from .service.heritage_update_service import HeritageUpdateService
from .service.msg_forwarding_service import SignedMessageWithReplyForwardingService
from .service.msg_forwarding_service import SignedMessageWithoutReplyForwardingService
from .service.node_service import EventualNodeServiceWithoutReply
from .service.node_service import ImmediateNodeServiceWithReply
from .service.obj_action_service import EventualObjectActionServiceWithoutReply
from .service.obj_action_service import ImmediateObjectActionServiceWithReply
from .service.obj_action_service import ImmediateObjectActionServiceWithoutReply
from .service.obj_search_permission_service import (
    ImmediateObjectSearchPermissionUpdateService,
)
from .service.obj_search_service import ImmediateObjectSearchService
from .service.repr_service import ReprService
from .service.resolve_pointer_type_service import ResolvePointerTypeService

# this generic type for Client bound by Client
ClientT = TypeVar("ClientT", bound=Client)


# TODO: Move but right now import loop prevents importing from the RequestMessage
class DuplicateRequestException(Exception):
    pass


class Node(AbstractNode):

    """
    Basic class for a syft node behavior, explicit purpose node will
    inherit this class (e.g., Device, Domain, Network, and VirtualMachine).

    Each node is identified by an id of type ID and a name of type string.
    """

    client_type = ClientT
    child_type_client_type = ClientT

    ChildT = TypeVar("ChildT", bound="Node")
    child_type = ChildT

    signing_key: Optional[SigningKey]
    verify_key: Optional[VerifyKey]

    def __init__(
        self,
        name: Optional[str] = None,
        network: Optional[Location] = None,
        domain: Optional[Location] = None,
        device: Optional[Location] = None,
        vm: Optional[Location] = None,
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
        db_path: Optional[str] = None,
    ):

        # The node has a name - it exists purely to help the
        # end user have some idea about what this node is in a human
        # readable form. It is not guaranteed to be unique (or to
        # really be anything for that matter).
        super().__init__(
            name=name, network=network, domain=domain, device=device, vm=vm
        )

        # Any object that needs to be stored on a node is stored here
        # More specifically, all collections of objects are found here
        # There should be NO COLLECTIONS stored as attributes directly
        # on a Node if there is a chance that the collections could
        # become quite numerous (or otherwise fill up RAM).
        # self.store is the elastic memory.

        if db_path is not None:
            try:
                self.store = DiskObjectStore(db_path=db_path)
                log = f"Opened DiskObjectStore at {db_path}."
                debug(log)
            except Exception as e:
                log = f"Failed to open DiskObjectStore at {db_path}. {e}"
                critical(log)
        else:
            self.store = MemoryStore()
            log = "Created MemoryStore."
            debug(log)

        # We need to register all the services once a node is created
        # On the off chance someone forgot to do this (super unlikely)
        # this flag exists to check it.
        self.services_registered = False

        # In order to be able to write generic services (in .service)
        # which can work for all node types, sometimes we need to have
        # a reference to what node type this node is. This attribute
        # provides that ability.
        self.node_type = type(self).__name__

        # ABOUT SERVICES AND MESSAGES

        # Each service corresponds to one or more message types which
        # the service processes. There are two kinds of messages, those
        # which require a reply and those which do not. Thus, there
        # are two kinds of services, service which generate a reply
        # and services which do not. It's important to distinguish
        # between them because:
        #
        # 1) services which do not return a reply
        # can typically be run on a more flexible time-table, whereas
        # services which require a reply often care about the latency
        # of the reply.
        #
        # 2) Services which do not return a reply aren't likely to leak
        # any information because no information is leaving. Thus, our
        # privacy/security concerns are more concentrated within service
        # which reply with some amount of information.

        # for messages which need a reply, this uses the type
        # of the message to look up the service which
        # addresses that message.
        self.immediate_msg_with_reply_router: Dict[
            Type[ImmediateSyftMessageWithReply], ImmediateNodeServiceWithReply
        ] = {}

        # for messages which don't lead to a reply, this uses
        # the type of the message to look up the service
        # which addresses that message
        self.immediate_msg_without_reply_router: Dict[
            Type[ImmediateSyftMessageWithoutReply], Any
        ] = {}

        # for messages which don't need to be run right now
        # and will not generate a reply.
        self.eventual_msg_without_reply_router: Dict[
            Type[EventualSyftMessageWithoutReply], EventualNodeServiceWithoutReply
        ] = {}

        # This is the list of services which all node support.
        # You can read more about them by reading their respective
        # class documentation.

        # TODO: Support ImmediateNodeServiceWithoutReply Parent Class
        # for services which run immediately and do not return a reply
        self.immediate_services_without_reply: List[Any] = []
        self.immediate_services_without_reply.append(ReprService)
        self.immediate_services_without_reply.append(HeritageUpdateService)
        self.immediate_services_without_reply.append(ChildNodeLifecycleService)
        self.immediate_services_without_reply.append(
            ImmediateObjectActionServiceWithoutReply
        )
        self.immediate_services_without_reply.append(
            ImmediateObjectSearchPermissionUpdateService
        )

        # TODO: Support ImmediateNodeServiceWithReply Parent Class
        # for services which run immediately and return a reply
        self.immediate_services_with_reply: List[Any] = []
        self.immediate_services_with_reply.append(ImmediateObjectActionServiceWithReply)
        self.immediate_services_with_reply.append(ImmediateObjectSearchService)
        self.immediate_services_with_reply.append(GetReprService)
        self.immediate_services_with_reply.append(ResolvePointerTypeService)

        # for services which can run at a later time and do not return a reply
        self.eventual_services_without_reply = list()
        self.eventual_services_without_reply.append(
            EventualObjectActionServiceWithoutReply
        )

        # This is a special service which cannot be listed in any
        # of the other services because it handles messages of all types.
        # Thus, it does not live in a message router since
        # routers only exist to decide which messages go to which
        # services, and they require that every message only correspond
        # to only one service type. If we have more messages like
        # these we'll make a special category for "services that
        # all messages are applied to" but for now this will do.

        self.signed_message_with_reply_forwarding_service = (
            SignedMessageWithReplyForwardingService()
        )

        self.signed_message_without_reply_forwarding_service = (
            SignedMessageWithoutReplyForwardingService()
        )

        # now we need to load the relevant frameworks onto the node
        self.lib_ast = lib_ast
        # The node needs to sign messages that it sends so that recipients know that it
        # comes from the node. In order to do that, the node needs to generate keys
        # for itself to sign and verify with.

        # create a signing key if one isn't provided
        if signing_key is None:
            self.signing_key = SigningKey.generate()
        else:
            self.signing_key = signing_key

        # if verify key isn't provided, get verify key from signing key
        if verify_key is None:
            self.verify_key = self.signing_key.verify_key
        else:
            self.verify_key = verify_key

        # PERMISSION REGISTRY:
        self.root_verify_key = self.verify_key  # TODO: CHANGE
        self.guest_verify_key_registry = set()
        self.admin_verify_key_registry = set()
        self.cpl_ofcr_verify_key_registry = set()
        self.in_memory_client_registry = {}
        # TODO: remove hacky signaling_msgs when SyftMessages become Storable.
        self.signaling_msgs = {}

        # For logging the number of messages received
        self.message_counter = 0

    @property
    def icon(self) -> str:
        return "📍"

    def get_client(self, routes: Optional[List[Route]] = None) -> ClientT:
        if not routes:
            conn_client = create_virtual_connection(node=self)
            solo = SoloRoute(destination=self.target_id, connection=conn_client)
            # inject name
            setattr(solo, "name", f"Route ({self.name} <-> {self.name} Client)")
            routes = [solo]

        return self.client_type(  # type: ignore
            name=self.name,
            routes=routes,
            network=self.network,
            domain=self.domain,
            device=self.device,
            vm=self.vm,
            signing_key=None,  # DO NOT PASS IN A SIGNING KEY!!! The client generates one.
            verify_key=None,  # DO NOT PASS IN A VERIFY KEY!!! The client generates one.
        )

    def get_root_client(self, routes: Optional[List[Route]] = None) -> ClientT:
        client: ClientT = self.get_client(routes=routes)
        self.root_verify_key = client.verify_key
        return client

    def get_metadata_for_client(self) -> Metadata:
        return Metadata(
            name=self.name if self.name else "", id=self.id, node=self.target_id
        )

    @property
    def known_nodes(self) -> List[Client]:
        """This is a property which returns a list of all known node
        by returning the clients we used to interact with them from
        the object store."""
        return list(self.in_memory_client_registry.values())

    @property
    def id(self) -> UID:
        traceback_and_raise(NotImplementedError)

    @property
    def known_child_nodes(self) -> List[Address]:
        debug(f"> {self.pprint} Getting known Children Nodes")
        if self.child_type_client_type is not None:
            return [
                client
                for client in self.in_memory_client_registry.values()
                if all(
                    [
                        self.network is None
                        or client.network is None
                        or self.network == client.network,
                        self.domain is None
                        or client.domain is None
                        or self.domain == client.domain,
                        self.device is None
                        or client.device is None
                        or self.device == client.device,
                        self.vm is None or client.vm is None or self.vm == client.vm,
                    ]
                )
            ]
        else:
            debug(f"> Node {self.pprint} has no children")
            return []

    def message_is_for_me(self, msg: Union[SyftMessage, SignedMessage]) -> bool:
        traceback_and_raise(NotImplementedError)

    def recv_immediate_msg_with_reply(
        self, msg: SignedImmediateSyftMessageWithReply
    ) -> SignedImmediateSyftMessageWithoutReply:
        # exceptions can be easily triggered which break any WebRTC loops
        # so we need to catch them here and respond with a special exception
        # message reply
        try:
            debug(
                f"> Received with Reply {msg.message.pprint} {msg.message.id} @ {self.pprint}"
            )
            # try to process message
            response = self.process_message(
                msg=msg, router=self.immediate_msg_with_reply_router
            )

        except Exception as e:
            error(e)
            public_exception: Exception
            if isinstance(e, AuthorizationException):
                private_log_msg = "An AuthorizationException has been triggered"
                public_exception = e
            else:
                private_log_msg = f"An {type(e)} has been triggered"  # dont send
                public_exception = UnknownPrivateException(
                    "UnknownPrivateException has been triggered."
                )
            try:
                # try printing a useful message
                private_log_msg += f" by {type(msg.message)} "
                private_log_msg += f"from {msg.message.reply_to}"  # type: ignore
            except Exception:
                error("Unable to format the private log message")
                pass
            # show the host what the real exception is
            error(private_log_msg)

            # send the public exception back
            response = ExceptionMessage(
                address=msg.message.reply_to,  # type: ignore
                msg_id_causing_exception=msg.message.id,
                exception_type=type(public_exception),
                exception_msg=str(public_exception),
            )

        # maybe I shouldn't have created process_message because it screws up
        # all the type inference.
        res_msg = response.sign(signing_key=self.signing_key)  # type: ignore
        output = (
            f"> {self.pprint} Signing {res_msg.pprint} with "
            + f"{self.key_emoji(key=self.signing_key.verify_key)}"  # type: ignore
        )
        debug(output)
        return res_msg

    def recv_immediate_msg_without_reply(
        self, msg: SignedImmediateSyftMessageWithoutReply
    ) -> None:
        debug(
            f"> Received without Reply {msg.message.pprint} {msg.message.id} @ {self.pprint}"
        )

        self.process_message(msg=msg, router=self.immediate_msg_without_reply_router)
        try:
            pass
        except Exception as e:
            error(f"Exception processing {msg.message}. {e}")
            # public_exception: Exception
            if isinstance(e, DuplicateRequestException):
                private_log_msg = "An DuplicateRequestException has been triggered"
                # public_exception = e
            else:
                private_log_msg = f"An {type(e)} has been triggered"  # dont send
                # public_exception = UnknownPrivateException(
                #     "UnknownPrivateException has been triggered."
                # )
            try:
                # try printing a useful message
                private_log_msg += f" by {type(msg.message)} "
                private_log_msg += f"from {msg.message.reply_to}"  # type: ignore
            except Exception:
                error("Unable to format the private log message")
                pass
            # show the host what the real exception is
            error(private_log_msg)

            # we still want to raise for now due to certain exceptions we expect
            # in tests
            if not isinstance(e, DuplicateRequestException):
                error(e)
                # TODO: A lot of tests are depending on this raise which seems bad
                traceback_and_raise(e)

            # TODO: finish code to send ExceptionMessage back
            # if isinstance(e, DuplicateRequestException):
            #     # we have a reply_to
            #     # send the public exception back
            #     response = ExceptionMessage(
            #         address=msg.message.reply_to,  # type: ignore
            #         msg_id_causing_exception=msg.message.id,
            #         exception_type=type(public_exception),
            #         exception_msg=str(public_exception),
            #     )
            #     res_msg = response.sign(signing_key=self.signing_key)  # type: ignore
            #     self.client.send_immediate_msg_with_reply.process(
            #         node=self,
            #         msg=res_msg,
            #     )
        return None

    def recv_eventual_msg_without_reply(
        self, msg: SignedEventualSyftMessageWithoutReply
    ) -> None:
        self.process_message(msg=msg, router=self.eventual_msg_without_reply_router)

    # TODO: Add SignedEventualSyftMessageWithoutReply and others
    def process_message(
        self, msg: SignedMessage, router: dict
    ) -> Union[SyftMessage, None]:

        self.message_counter += 1

        debug(f"> Processing 📨 {msg.pprint} @ {self.pprint} {msg.message}")
        if self.message_is_for_me(msg=msg):
            debug(
                f"> Recipient Found {msg.pprint}{msg.address.target_emoji()} == {self.pprint}"
            )
            # Process Message here
            if not msg.is_valid:
                error(f"Message is not valid. {msg}")
                traceback_and_raise(Exception("Message is not valid."))

            try:  # we use try/except here because it's marginally faster in Python
                service = router[type(msg.message)]
            except KeyError as e:
                log = (
                    f"The node {self.id} of type {type(self)} cannot process messages of type "
                    + f"{type(msg.message)} because there is no service running to process it."
                    + f"{e}"
                )
                error(log)
                self.ensure_services_have_been_registered_error_if_not()
                traceback_and_raise(KeyError(log))

            result = service.process(
                node=self,
                msg=msg.message,
                verify_key=msg.verify_key,
            )
            return result

        else:
            debug(
                f"> Recipient Not Found ↪️ {msg.pprint}{msg.address.target_emoji()} != {self.pprint}"
            )
            # Forward message onwards
            if issubclass(type(msg), SignedImmediateSyftMessageWithReply):
                return self.signed_message_with_reply_forwarding_service.process(
                    node=self,
                    msg=msg,  # type: ignore
                )
            if issubclass(type(msg), SignedImmediateSyftMessageWithoutReply):
                return self.signed_message_without_reply_forwarding_service.process(
                    node=self,
                    msg=msg,  # type: ignore
                )
        return None

    def ensure_services_have_been_registered_error_if_not(self) -> None:
        if not self.services_registered:
            traceback_and_raise(
                Exception(
                    "Please call _register_services on node. This seems to have"
                    "been skipped for some reason."
                )
            )

    def _register_services(self) -> None:
        """In this method, we set each message type to the appropriate
        service for this node. It's important to note that one message type
        cannot map to multiple services on any given node type. If you want to
        send information to a different service, create a new message type for that
        service. Put another way, a service can have multiple message types which
        correspond to it, but each message type can only have one service (per node
        subclass) which corresponds to it."""

        for isr in self.immediate_services_with_reply:
            # Create a single instance of the service to cache in the router corresponding
            # to one or more message types.
            isr_instance = isr()
            for handler_type in isr.message_handler_types():
                # for each explicitly supported type, add it to the router
                self.immediate_msg_with_reply_router[handler_type] = isr_instance

                # for all sub-classes of the explicitly supported type, add them
                # to the router as well.
                for handler_type_subclass in get_subclasses(obj_type=handler_type):
                    self.immediate_msg_with_reply_router[
                        handler_type_subclass
                    ] = isr_instance

        for iswr in self.immediate_services_without_reply:
            # Create a single instance of the service to cache in the router corresponding
            # to one or more message types.
            iswr_instance = iswr()
            for handler_type in iswr.message_handler_types():

                # for each explicitly supported type, add it to the router
                self.immediate_msg_without_reply_router[handler_type] = iswr_instance

                # for all sub-classes of the explicitly supported type, add them
                # to the router as well.
                for handler_type_subclass in get_subclasses(obj_type=handler_type):
                    self.immediate_msg_without_reply_router[
                        handler_type_subclass
                    ] = iswr_instance

        for eswr in self.eventual_services_without_reply:
            # Create a single instance of the service to cache in the router corresponding
            # to one or more message types.
            eswr_instance = eswr()
            for handler_type in eswr.message_handler_types():

                # for each explicitly supported type, add it to the router
                self.eventual_msg_without_reply_router[handler_type] = eswr_instance

                # for all sub-classes of the explicitly supported type, add them
                # to the router as well.
                for handler_type_subclass in get_subclasses(obj_type=handler_type):
                    self.eventual_msg_without_reply_router[
                        handler_type_subclass
                    ] = eswr_instance

        # Set the services_registered flag to true so that we know that all services
        # have been properly registered. This mostly exists because someone might
        # accidentally delete (forget to call) this method inside the __init__ function
        # of a sub-class of Node.
        self.services_registered = True

    def __repr__(self) -> str:
        no_dash = str(self.id).replace("-", "")
        return f"{self.node_type}: {self.name}: {no_dash}"



# syft absolute
import syft as sy
from syft.lib.python.string import String
from syft.proto.lib.python.string_pb2 import String as String_PB


def test_string_serde() -> None:
    syft_string = String("Hello OpenMined")

    serialized = syft_string._object2proto()

    assert isinstance(serialized, String_PB)

    deserialized = String._proto2object(proto=serialized)

    assert isinstance(deserialized, String)
    assert deserialized.id == syft_string.id


def test_string_send(client: sy.VirtualMachineClient) -> None:
    syft_string = String("Hello OpenMined!")
    ptr = syft_string.send(client)

    # Check pointer type
    assert ptr.__class__.__name__ == "StringPointer"

    # Check that we can get back the object
    res = ptr.get()
    assert res == syft_string



# stdlib
from typing import List
from typing import Optional

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
from typing_extensions import final

# syft relative
from ....logger import traceback_and_raise
from ...common.uid import UID
from ...io.location import Location
from ...io.location import SpecificLocation
from ...io.route import Route
from ..common.client import Client


@final
class DeviceClient(Client):

    device: SpecificLocation  # redefine the type of self.device to not be optional

    def __init__(
        self,
        name: Optional[str],
        routes: List[Route],
        device: SpecificLocation,
        network: Optional[Location] = None,
        domain: Optional[Location] = None,
        vm: Optional[Location] = None,
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
    ):
        super().__init__(
            name=name,
            routes=routes,
            network=network,
            domain=domain,
            device=device,
            vm=vm,
            signing_key=signing_key,
            verify_key=verify_key,
        )

        self.post_init()

    # def create_vm(self, name:str):
    #
    #     # Step 1: create a message which will request for a VM to be created.
    #     # we can set route=None because we know the message is just going directly
    #     # to the device.
    #     msg = CreateVirtualMachineMessage(vm_name=name, address=self.node_address)
    #
    #     # Step 2: Send the message to the device this client points to.
    #     # Receive a reply_msg and save it in a variable.
    #     reply_msg = self.send_msg(msg=msg)
    #
    #     # Step 3: Unpack the vm client from the reply message
    #     vm_client = reply_msg.client
    #
    #     # Step 4: Return the client object directly. We assume that the
    #     # client object knows how to send message to the VM no matter
    #     # where it exists. Aka, we don't need to save the client in any
    #     # particular kind of context to know how to use it. We should be
    #     # able to "just use it" anywhere and it should "just work"... finding
    #     # the appropriate VM.
    #     return vm_client

    @property
    def id(self) -> UID:
        """This client points to a vm, this returns the id of that vm."""
        return self.device.id

    @property
    def vm(self) -> Optional[Location]:
        """This client points to a node, if that node lives within a vm
        or is a vm itself, this property will return the Location of that vm
        if it is known by the client."""

        return super().vm

    @vm.setter
    def vm(self, new_vm: Location) -> Optional[Location]:
        """This client points to a node, if that node lives within a vm
        or is a vm itself and we learn the Location of that vm, this setter
        allows us to save the Location of that vm for use later. We use a getter
        (@property) and setter (@set) explicitly because we want all clients
        to efficiently save an address object for use when sending messages to their
        target. That address object will include this information if it is available"""

        traceback_and_raise(
            Exception("This client points to a Device, you don't need a VM Location.")
        )

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self.device}>"



# third party
import pytest

# syft absolute
import syft as sy


@pytest.mark.vendor(lib="zksk")
def test_secret_serde() -> None:
    vm = sy.VirtualMachine()
    client = vm.get_root_client()

    # third party
    import zksk as zk

    sy.load("zksk")

    r = zk.Secret(zk.utils.get_random_num(bits=128))
    r_ptr = r.send(client)
    r2 = r_ptr.get()

    assert r == r2



# syft relative
from .device_type import DeviceType
from .specs.cpu import CPU
from .specs.cpu import CpuArchitectureTypes
from .specs.provider import Provider
from .specs.storage import Drive
from .specs.storage import DriveType
from .specs.storage import Storage

mbp_storage = Storage([Drive(name="/", storage=256000, drive_type=DriveType.SSD)])
mbp_cpu = CPU(
    num_cores=6,
    num_threads_per_core=1,
    clock_speed=2.2,
    architecture=CpuArchitectureTypes.x86_64,
)
unknown_device = DeviceType(
    name="Unknown Device",
    provider=Provider.USER_OWNED,
    price_per_hour=0,
    memory=32768,
    bare_metal=True,
    storage=mbp_storage,
    cpu=mbp_cpu,
    gpu=None,
    network=None,
    spot_mode_supported=False,
    price_per_hour_spot=0,
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/domain/service/request_handler_message.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.lib.python import dict_pb2 as proto_dot_lib_dot_python_dot_dict__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/core/node/domain/service/request_handler_message.proto",
    package="syft.core.node.domain.service",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n<proto/core/node/domain/service/request_handler_message.proto\x12\x1dsyft.core.node.domain.service\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto\x1a\x1bproto/lib/python/dict.proto"\xa2\x01\n\x1bUpdateRequestHandlerMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12&\n\x07handler\x18\x03 \x01(\x0b\x32\x15.syft.lib.python.Dict\x12\x0c\n\x04keep\x18\x04 \x01(\x08"\x96\x01\n\x1cGetAllRequestHandlersMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\x9e\x01\n$GetAllRequestHandlersResponseMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08handlers\x18\x03 \x03(\x0b\x32\x15.syft.lib.python.Dictb\x06proto3',
    dependencies=[
        proto_dot_core_dot_common_dot_common__object__pb2.DESCRIPTOR,
        proto_dot_core_dot_io_dot_address__pb2.DESCRIPTOR,
        proto_dot_lib_dot_python_dot_dict__pb2.DESCRIPTOR,
    ],
)


_UPDATEREQUESTHANDLERMESSAGE = _descriptor.Descriptor(
    name="UpdateRequestHandlerMessage",
    full_name="syft.core.node.domain.service.UpdateRequestHandlerMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.core.node.domain.service.UpdateRequestHandlerMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.core.node.domain.service.UpdateRequestHandlerMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="handler",
            full_name="syft.core.node.domain.service.UpdateRequestHandlerMessage.handler",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="keep",
            full_name="syft.core.node.domain.service.UpdateRequestHandlerMessage.keep",
            index=3,
            number=4,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=193,
    serialized_end=355,
)


_GETALLREQUESTHANDLERSMESSAGE = _descriptor.Descriptor(
    name="GetAllRequestHandlersMessage",
    full_name="syft.core.node.domain.service.GetAllRequestHandlersMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.core.node.domain.service.GetAllRequestHandlersMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.core.node.domain.service.GetAllRequestHandlersMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.core.node.domain.service.GetAllRequestHandlersMessage.reply_to",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=358,
    serialized_end=508,
)


_GETALLREQUESTHANDLERSRESPONSEMESSAGE = _descriptor.Descriptor(
    name="GetAllRequestHandlersResponseMessage",
    full_name="syft.core.node.domain.service.GetAllRequestHandlersResponseMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.core.node.domain.service.GetAllRequestHandlersResponseMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.core.node.domain.service.GetAllRequestHandlersResponseMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="handlers",
            full_name="syft.core.node.domain.service.GetAllRequestHandlersResponseMessage.handlers",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=511,
    serialized_end=669,
)

_UPDATEREQUESTHANDLERMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_UPDATEREQUESTHANDLERMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_UPDATEREQUESTHANDLERMESSAGE.fields_by_name[
    "handler"
].message_type = proto_dot_lib_dot_python_dot_dict__pb2._DICT
_GETALLREQUESTHANDLERSMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_GETALLREQUESTHANDLERSMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETALLREQUESTHANDLERSMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETALLREQUESTHANDLERSRESPONSEMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_GETALLREQUESTHANDLERSRESPONSEMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETALLREQUESTHANDLERSRESPONSEMESSAGE.fields_by_name[
    "handlers"
].message_type = proto_dot_lib_dot_python_dot_dict__pb2._DICT
DESCRIPTOR.message_types_by_name[
    "UpdateRequestHandlerMessage"
] = _UPDATEREQUESTHANDLERMESSAGE
DESCRIPTOR.message_types_by_name[
    "GetAllRequestHandlersMessage"
] = _GETALLREQUESTHANDLERSMESSAGE
DESCRIPTOR.message_types_by_name[
    "GetAllRequestHandlersResponseMessage"
] = _GETALLREQUESTHANDLERSRESPONSEMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

UpdateRequestHandlerMessage = _reflection.GeneratedProtocolMessageType(
    "UpdateRequestHandlerMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATEREQUESTHANDLERMESSAGE,
        "__module__": "proto.core.node.domain.service.request_handler_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.domain.service.UpdateRequestHandlerMessage)
    },
)
_sym_db.RegisterMessage(UpdateRequestHandlerMessage)

GetAllRequestHandlersMessage = _reflection.GeneratedProtocolMessageType(
    "GetAllRequestHandlersMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETALLREQUESTHANDLERSMESSAGE,
        "__module__": "proto.core.node.domain.service.request_handler_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.domain.service.GetAllRequestHandlersMessage)
    },
)
_sym_db.RegisterMessage(GetAllRequestHandlersMessage)

GetAllRequestHandlersResponseMessage = _reflection.GeneratedProtocolMessageType(
    "GetAllRequestHandlersResponseMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETALLREQUESTHANDLERSRESPONSEMESSAGE,
        "__module__": "proto.core.node.domain.service.request_handler_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.domain.service.GetAllRequestHandlersResponseMessage)
    },
)
_sym_db.RegisterMessage(GetAllRequestHandlersResponseMessage)


# @@protoc_insertion_point(module_scope)



# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# third party
import opacus

# syft relative
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..util import generic_update_ast

LIB_NAME = "opacus"
PACKAGE_SUPPORT = {
    "lib": LIB_NAME,
    "torch": {"min_version": "1.6.0"},
    "python": {"min_version": (3, 6, 9), "max_version": (3, 9, 99)},
}


def create_ast(client: TypeAny = None) -> Globals:
    ast = Globals(client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [
        ("opacus", opacus),
        ("opacus.privacy_engine", opacus.privacy_engine),
    ]
    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        (
            "opacus.privacy_engine.PrivacyEngine",
            "opacus.privacy_engine.PrivacyEngine",
            opacus.privacy_engine.PrivacyEngine,
        ),
    ]

    methods = [
        (
            "opacus.privacy_engine.PrivacyEngine.to",
            "opacus.privacy_engine.PrivacyEngine",
        ),
        ("opacus.privacy_engine.PrivacyEngine.step", "syft.lib.python._SyNone"),
        ("opacus.privacy_engine.PrivacyEngine.zero_grad", "syft.lib.python._SyNone"),
        ("opacus.privacy_engine.PrivacyEngine.zero_grad", "syft.lib.python._SyNone"),
        (
            "opacus.privacy_engine.PrivacyEngine.get_privacy_spent",
            "syft.lib.python.Tuple",
        ),
        ("opacus.privacy_engine.PrivacyEngine.attach", "syft.lib.python._SyNone"),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)



# stdlib
from os import listdir
from os.path import join

# third party
from PIL import Image
import torch.utils.data as data


def is_image_file(filename):
    return any(filename.endswith(extension) for extension in [".png", ".jpg", ".jpeg"])


def load_img(filepath):
    img = Image.open(filepath).convert("YCbCr")
    y, _, _ = img.split()
    return y


class DatasetFromFolder(data.Dataset):
    def __init__(self, image_dir, input_transform=None, target_transform=None):
        super(DatasetFromFolder, self).__init__()
        self.image_filenames = [
            join(image_dir, x) for x in listdir(image_dir) if is_image_file(x)
        ]

        self.input_transform = input_transform
        self.target_transform = target_transform

    def __getitem__(self, index):
        input = load_img(self.image_filenames[index])
        target = input.copy()
        if self.input_transform:
            input = self.input_transform(input)
        if self.target_transform:
            target = self.target_transform(target)

        return input, target

    def __len__(self):
        return len(self.image_filenames)



# stdlib
import inspect
from types import ModuleType
from typing import Callable
from typing import Optional
from typing import Union
from typing import Union as TypeUnion

# syft relative
from ..ast.globals import Globals
from ..core.node.abstract.node import AbstractNodeClient

# this gets called on global ast as well as clients
# anything which wants to have its ast updated and has an add_attr method


def generic_update_ast(
    lib_name: str,
    create_ast: Callable,
    ast_or_client: TypeUnion[Globals, AbstractNodeClient],
) -> None:
    if isinstance(ast_or_client, Globals):
        ast = ast_or_client
        new_lib_ast = create_ast(None)
        ast.add_attr(attr_name=lib_name, attr=new_lib_ast.attrs[lib_name])
    elif isinstance(ast_or_client, AbstractNodeClient):
        client = ast_or_client
        new_lib_ast = create_ast(client)
        client.lib_ast.attrs[lib_name] = new_lib_ast.attrs[lib_name]
        setattr(client, lib_name, new_lib_ast.attrs[lib_name])
    else:
        raise ValueError(
            f"Expected param of type (Globals, AbstractNodeClient), but got {type(ast_or_client)}"
        )


def is_static_method(klass: type, attr: str) -> bool:
    """Test if a value of a class is static method.

    Example:

    .. code-block::

        class MyClass(object):
            @staticmethod
            def method():
                ...

    Args:
        klass (type): the class on which we want to check whether the method is statically implemented
        attr (str): the name of the method we want to check.

    Returns:
        bool: whether or not a method named <attr> is a static method of class <klass>
    """

    if not inspect.isclass(klass):
        return False

    if hasattr(klass, attr):
        value = getattr(klass, attr)
    else:
        return False

    if getattr(klass, attr) != value:
        raise AttributeError("Method don't match")

    for cls in inspect.getmro(klass):
        if inspect.isroutine(value):
            if attr in cls.__dict__:
                bound_value = cls.__dict__[attr]
                if isinstance(bound_value, staticmethod):
                    return True
    return False


def copy_static_methods(from_class: type, to_class: type) -> None:
    """Copies all static methods from one class to another class

    This utility was initialized during the creation of the Constructor for PyTorch's "th.Tensor" class. Since we
    replace each original constructor (th.Tensor) with on we implement (torch.UppercaseTensorConstructor), we also
    need to make sure that our new constructor has any static methods which were previously stored on th.Tensor.
    Otherwise, the library might look for them there, not find them, and then trigger an error.

    Args:
        from_class (Type): the class on which we look for static methods co copy
        to_class (Type): the class onto which we copy all static methods found in <from_class>

    """
    # there are no static methods if from_class itself is not a type (sometimes funcs get passed in)

    for attr in dir(from_class):
        if is_static_method(klass=from_class, attr=attr):
            setattr(to_class, attr, getattr(from_class, attr))


def get_original_constructor_name(object_name: str) -> str:
    """Generate name for original constructor

    For each custom constructor, we move the original constructor to a consistent location relative to
    the original constructor so that each custom constructor automatically knows where to find the original
    method it is overloading. Namely, we move the original constructor to a different attr within the same
    module as the original constructor. This method specifies the naming convention that we use to name
    the original constructor when it is moved.

      Args:
          object_name (str): the original constructor's original name
    """

    return f"original_{object_name}"


def replace_classes_in_module(
    module: ModuleType,
    from_class: Callable,
    to_class: Callable,
    ignore_prefix: Optional[str] = None,
) -> None:
    """Recursively replace occurrence of `from_class` to `to_class` inside module.

    For example, when syft replaces torch.nn.parameter.Parameter constructor,
    there's also need to replace the same constructor in other modules that has already
    imported it.

    Args:
        module (ModuleType): top-level module to traverse
        from_class: Original constructor
        to_class: syft's ObjectConstructor
    """
    visited_modules = []

    def recursive_update(
        module: ModuleType, attr_name: Union[str, None] = None
    ) -> None:

        # check if we need to skip this attribute to preserve our unmodified
        # original copy
        if (
            attr_name is not None
            and ignore_prefix is not None
            and attr_name.startswith(ignore_prefix)
        ):
            # found an attr that should be skipped so lets return
            return
        attr = getattr(module, attr_name) if isinstance(attr_name, str) else module
        if isinstance(attr, ModuleType) and attr not in visited_modules:
            visited_modules.append(attr)
            for child_attr_name in dir(attr):
                recursive_update(attr, child_attr_name)
        elif (
            isinstance(attr_name, str) and inspect.isclass(attr) and attr is from_class
        ):
            setattr(module, attr_name, to_class)

    recursive_update(module)


def full_name_with_qualname(klass: type) -> str:
    return f"{klass.__module__}.{klass.__qualname__}"


def full_name_with_name(klass: type) -> str:
    return f"{klass.__module__}.{klass.__name__}"



"""
Warm the MNIST cache: https://github.com/pytorch/vision/issues/1938
"""
# third party
from six.moves import urllib
from torchvision import datasets

# syft absolute
from syft.util import get_root_data_path

real_user_agent = (
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 "
    "(KHTML, like Gecko) Chrome/88.0.4324.192 Safari/537.36"
)
opener = urllib.request.build_opener()
opener.addheaders = [("User-agent", real_user_agent)]
urllib.request.install_opener(opener)

# https://github.com/pytorch/vision/issues/3549
datasets.MNIST.resources = [
    (
        "https://ossci-datasets.s3.amazonaws.com/mnist/train-images-idx3-ubyte.gz",
        "f68b3c2dcbeaaa9fbdd348bbdeb94873",
    ),
    (
        "https://ossci-datasets.s3.amazonaws.com/mnist/train-labels-idx1-ubyte.gz",
        "d53e105ee54ea40749a09fcbcd1e9432",
    ),
    (
        "https://ossci-datasets.s3.amazonaws.com/mnist/t10k-images-idx3-ubyte.gz",
        "9fb629c4189551a2d022fa330f9573f3",
    ),
    (
        "https://ossci-datasets.s3.amazonaws.com/mnist/t10k-labels-idx1-ubyte.gz",
        "ec29112dd5afa0611ce80d1b7f02629c",
    ),
]


datasets.MNIST(get_root_data_path(), train=True, download=True)
datasets.MNIST(get_root_data_path(), train=False, download=True)



# stdlib
import os

# third party
import torch as th
from torch.utils.data import Dataset

# syft absolute
import syft as sy
from syft.core.remote_dataloader import RemoteDataLoader
from syft.core.remote_dataloader import RemoteDataset


class ExampleDataset(Dataset):
    def __init__(self, ten: th.Tensor):
        self.ten = ten

    def __len__(self) -> int:
        return self.ten.shape[0]

    def __getitem__(self, i: int) -> th.Tensor:
        return self.ten[i]


ten = th.rand((1000, 4))
ds = ExampleDataset(ten)


def test_remote_dataset(
    node: sy.VirtualMachine, client: sy.VirtualMachineClient
) -> None:
    th.save(ds, "ds.pt")

    rds = RemoteDataset(path="ds.pt", data_type="torch_tensor")
    rds_ptr = rds.send(client)
    rds_ptr.load_dataset()

    assert rds_ptr.len().get() == 1000
    for tp in rds_ptr:
        assert isinstance(tp.get(), th.Tensor)

    os.system("rm ds.pt")


def test_remote_dataloader(root_client: sy.VirtualMachineClient) -> None:
    th.save(ds, "ds.pt")

    rds = RemoteDataset(path="ds.pt", data_type="torch_tensor")
    rdl = RemoteDataLoader(remote_dataset=rds, batch_size=4)
    rdl_ptr = rdl.send(root_client)

    rdl_ptr.load_dataset()
    rdl_ptr.create_dataloader()

    assert rdl_ptr.len().get() == 250
    for tp in rdl_ptr:
        assert isinstance(tp.get(), th.Tensor)
    os.system("rm ds.pt")



# stdlib
import sys

# third party
from packaging import version
import tenseal as ts

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...logger import info
from ...logger import traceback_and_raise
from ...proto.util.vendor_bytes_pb2 import VendorBytes as VendorBytes_PB
from ...util import get_fully_qualified_name
from ..util import full_name_with_name


def context_object2proto(obj: object) -> VendorBytes_PB:
    proto = VendorBytes_PB()
    proto.obj_type = full_name_with_name(klass=obj._sy_serializable_wrapper_type)  # type: ignore
    proto.vendor_lib = "tenseal"
    proto.vendor_lib_version = ts.__version__
    proto.content = obj.serialize(save_secret_key=True)  # type: ignore

    return proto


def context_proto2object(proto: VendorBytes_PB) -> ts.Context:
    vendor_lib = proto.vendor_lib
    lib_version = version.parse(proto.vendor_lib_version)

    if vendor_lib not in sys.modules:
        traceback_and_raise(
            Exception(f"{vendor_lib} version: {proto.vendor_lib_version} is required")
        )
    else:
        if lib_version > version.parse(ts.__version__):
            log = f"Warning {lib_version} > local imported version {ts.__version__}"
            info(log)

    return ts.context_from(proto.content, n_threads=1)


GenerateWrapper(
    wrapped_type=ts.Context,
    import_path="tenseal.Context",
    protobuf_scheme=VendorBytes_PB,
    type_object2proto=context_object2proto,
    type_proto2object=context_proto2object,
)


def schemetype_object2proto(obj: object) -> VendorBytes_PB:
    proto = VendorBytes_PB()
    proto.obj_type = get_fully_qualified_name(obj=obj)
    proto.vendor_lib = "tenseal"
    proto.vendor_lib_version = ts.__version__
    proto.content = str.encode(obj.name)  # type: ignore

    return proto


def schemetype_proto2object(proto: VendorBytes_PB) -> ts.SCHEME_TYPE:
    vendor_lib = proto.vendor_lib
    lib_version = version.parse(proto.vendor_lib_version)

    if vendor_lib not in sys.modules:
        traceback_and_raise(
            Exception(f"{vendor_lib} version: {proto.vendor_lib_version} is required")
        )
    else:
        if lib_version > version.parse(ts.__version__):
            log = f"Warning {lib_version} > local imported version {ts.__version__}"
            info(log)

    return ts.SCHEME_TYPE[proto.content.decode()]


GenerateWrapper(
    wrapped_type=ts.SCHEME_TYPE,
    import_path="tenseal.SCHEME_TYPE",
    protobuf_scheme=VendorBytes_PB,
    type_object2proto=schemetype_object2proto,
    type_proto2object=schemetype_proto2object,
)



# syft absolute
from syft.core.common.message import SignedEventualSyftMessageWithoutReply
from syft.core.common.message import SignedImmediateSyftMessageWithReply
from syft.core.common.message import SignedImmediateSyftMessageWithoutReply
from syft.core.io.virtual import VirtualClientConnection
from syft.core.io.virtual import VirtualServerConnection
from syft.core.io.virtual import create_virtual_connection

# syft relative
from .utils_test import MockNode
from .utils_test import construct_dummy_message

# --------------------- HELPER FUNCTION ---------------------


def test_create_virtual_connection() -> None:
    """
    Test that create_virtual_connection function returns VirtualClientConnection.
    """
    client = create_virtual_connection(node=MockNode())
    assert isinstance(client, VirtualClientConnection)


# --------------------- INITIALIZATION ---------------------


def test_virtual_server_connection_init() -> None:
    """
    Test that VirtualServerConnection uses the AbstractNode
    passed into its constructor.
    """
    node = MockNode()
    server = VirtualServerConnection(node=node)

    assert server.node == node


def test_virtual_client_connection_init() -> None:
    """
    Test that VirtualClientConnection uses the VirtualServerConnection
    passed into its constructor.
    """
    node = MockNode()
    server = VirtualServerConnection(node=node)
    client = VirtualClientConnection(server=server)
    assert client.server == server


# --------------------- CLASS METHODS ---------------------


def test_virtual_server_connection_recv_immedaite_msg_with_reply() -> None:
    """Test that VirtualServerConnection.recv_immedaite_msg_with_reply works."""
    node = MockNode()
    server = VirtualServerConnection(node=node)
    msg = construct_dummy_message(SignedImmediateSyftMessageWithReply)

    assert isinstance(
        server.recv_immediate_msg_with_reply(msg=msg),
        SignedImmediateSyftMessageWithoutReply,
    )


def test_virtual_server_connection_recv_immedaite_msg_without_reply() -> None:
    """Test that VirtualServerConnection.recv_immedaite_msg_without_reply works."""
    node = MockNode()
    server = VirtualServerConnection(node=node)
    msg = construct_dummy_message(SignedImmediateSyftMessageWithoutReply)

    assert not server.recv_immediate_msg_without_reply(msg=msg)


def test_virtual_server_connection_recv_eventual_msg_without_reply() -> None:
    """
    Test that VirtualServerConnection.recv_eventual_msg_with_reply raises NotImplementedError.
    """
    node = MockNode()
    server = VirtualServerConnection(node=node)
    msg = construct_dummy_message(SignedEventualSyftMessageWithoutReply)

    assert server.recv_eventual_msg_without_reply(msg=msg) is None


def test_virtual_client_connection_send_immediate_msg_without_reply() -> None:
    """
    Test that VirtualClientConnection.send_immediate_msg_without_reply works.
    """
    node = MockNode()
    server = VirtualServerConnection(node=node)
    client = VirtualClientConnection(server=server)
    msg = construct_dummy_message(SignedImmediateSyftMessageWithoutReply)

    assert client.send_immediate_msg_without_reply(msg=msg) is None


def test_virtual_client_connection_send_immediate_msg_with_reply() -> None:
    """
    Test that VirtualClientConnection.send_immediate_msg_with_reply works.
    """
    node = MockNode()
    server = VirtualServerConnection(node=node)
    client = VirtualClientConnection(server=server)
    msg = construct_dummy_message(SignedImmediateSyftMessageWithReply)

    assert isinstance(
        client.send_immediate_msg_with_reply(msg=msg),
        SignedImmediateSyftMessageWithoutReply,
    )


def test_virtual_client_connection_send_eventual_msg_without_reply() -> None:
    """
    Test that VirtualClientConnection.send_eventual_msg_with_reply raises NotImplementedError.
    """
    node = MockNode()
    server = VirtualServerConnection(node=node)
    client = VirtualClientConnection(server=server)
    msg = construct_dummy_message(SignedEventualSyftMessageWithoutReply)

    assert client.send_eventual_msg_without_reply(msg=msg) is None



# stdlib
from typing import Any
from typing import Dict

# syft absolute
import syft as sy
from syft import serialize
from syft.core.io.address import Address
from syft.grid.messages.infra_messages import CreateWorkerMessage
from syft.grid.messages.infra_messages import CreateWorkerResponse
from syft.grid.messages.infra_messages import DeleteWorkerMessage
from syft.grid.messages.infra_messages import DeleteWorkerResponse
from syft.grid.messages.infra_messages import GetWorkerMessage
from syft.grid.messages.infra_messages import GetWorkerResponse
from syft.grid.messages.infra_messages import GetWorkersMessage
from syft.grid.messages.infra_messages import GetWorkersResponse
from syft.grid.messages.infra_messages import UpdateWorkerMessage
from syft.grid.messages.infra_messages import UpdateWorkerResponse


def test_create_worker_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    request_content = {
        "settings": {
            "instance-size": "t4g.medium",
            "vCPU": "2",
            "network-bandwith": "5Gbps",
            "vGPU": True,
        }
    }
    msg = CreateWorkerMessage(
        address=target,
        content=request_content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_create_worker_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {"msg": "Worker Environment Created Successfully!"}
    msg = CreateWorkerResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_delete_worker_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    request_content = {"worker_id": "f2a6as5d16fasd"}
    msg = DeleteWorkerMessage(
        address=target,
        content=request_content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_delete_worker_response_serde() -> None:
    target = Address(name="Alice")

    content = {"msg": "Worker Environment deleted Successfully!"}
    msg = DeleteWorkerResponse(
        status_code=200,
        address=target,
        content=content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_update_worker_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content = {
        "worker-id": "eqw9e4a5d846",
        "settings": {
            "instance-size": "t4g.large",
            "vCPU": "2",
            "network-bandwith": "5Gbps",
            "vGPU": True,
        },
    }
    msg = UpdateWorkerMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_update_worker_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {"msg": "Worker Environment updated successfully!"}
    msg = UpdateWorkerResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_worker_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content = {"worker-id": "eqw9e4a5d846"}
    msg = GetWorkerMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_worker_response_serde() -> None:
    target = Address(name="Alice")

    content = {
        "worker-id": "eqw9e4a5d846",
        "environment-name": "Heart Diseases Environment",
        "owner": "user-id7",
        "deployment-date": "05/12/2021",
    }

    msg = GetWorkerResponse(
        address=target,
        status_code=200,
        content=content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_all_workers_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content: Dict[Any, Any] = {}
    msg = GetWorkersMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_all_workers_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {
        "workers": {
            "626sadaf631": {
                "environment-name": "Heart Diseases Environment",
                "owner": "user-id7",
                "deployment-date": "05/12/2021",
            },
            "a84ew64wq6e": {
                "worker-id": "eqw9e4a5d846",
                "environment-name": "Brain Diseases Environment",
                "owner": "user-id8",
                "deployment-date": "15/12/2021",
            },
        }
    }

    msg = GetWorkersResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2



# stdlib
from typing import Any as TypeAny

# syft relative
from .types import SyPrimitiveRet


class Any:
    def __iter__(self) -> "Any":
        return self

    def __next__(self) -> SyPrimitiveRet:
        return self.__next__()

    def __add__(self, other: TypeAny) -> SyPrimitiveRet:
        return self + other

    def __radd__(self, other: TypeAny) -> SyPrimitiveRet:
        return other + self

    def __truediv__(self, other: TypeAny) -> SyPrimitiveRet:
        return self / other

    def __rtruediv__(self, other: TypeAny) -> SyPrimitiveRet:
        return other / self

    def __floordiv__(self, other: TypeAny) -> SyPrimitiveRet:
        return self / other

    def __rfloordiv__(self, other: TypeAny) -> SyPrimitiveRet:
        return self / other

    def __mul__(self, other: TypeAny) -> SyPrimitiveRet:
        return self * other

    def __rmul__(self, other: TypeAny) -> SyPrimitiveRet:
        return other * self

    def __sub__(self, other: TypeAny) -> SyPrimitiveRet:
        return self - other

    def __rsub__(self, other: TypeAny) -> SyPrimitiveRet:
        return other - self

    def __len__(self) -> SyPrimitiveRet:
        return self.__len__()



"""The purpose of this application is to allow us to dev and test PySyft
functionality on an actual local network. This is NOT meant to be run in
production (that's the *actual* grid's job)."""

# third party
from flask import Flask

# syft absolute
from syft.core.node.network.network import Network

app = Flask(__name__)
network = Network(name="ucsf-net")


@app.route("/")
def hello_world() -> str:
    return "A network:" + str(network.name)


def run() -> None:
    app.run()


run()



# stdlib
from typing import List

# third party
import pytest

# syft absolute
import syft as sy


@pytest.mark.vendor(lib="numpy")
def test_remote_numpy_array(root_client: sy.VirtualMachineClient) -> None:
    # third party
    import numpy as np

    # syft absolute
    from syft.lib.numpy.array import SUPPORTED_BOOL_TYPES
    from syft.lib.numpy.array import SUPPORTED_DTYPES
    from syft.lib.numpy.array import SUPPORTED_FLOAT_TYPES
    from syft.lib.numpy.array import SUPPORTED_INT_TYPES

    sy.load("numpy")

    test_arrays: List[np.ndarray] = []
    for dtype in SUPPORTED_DTYPES:

        # test their bounds
        if dtype in SUPPORTED_BOOL_TYPES:
            lower = False
            upper = True
            mid = False
        elif dtype in SUPPORTED_INT_TYPES:
            bounds = np.iinfo(dtype)
            lower = bounds.min
            upper = bounds.max
            mid = upper + lower  # type: ignore
            if lower == 0:
                mid = round(mid / 2)  # type: ignore
        elif dtype in SUPPORTED_FLOAT_TYPES:
            bounds = np.finfo(dtype)
            lower = bounds.min
            upper = bounds.max
            mid = upper + lower  # type: ignore

        test_arrays.append(np.array([lower, mid, upper], dtype=dtype))

    for test_array in test_arrays:
        remote_array = test_array.send(root_client)
        received_array = remote_array.get()

        assert all(test_array == received_array)
        assert test_array.dtype == received_array.dtype



# syft absolute
import syft as sy
from syft import serialize
from syft.core.common.uid import UID
from syft.core.io.address import Address
from syft.core.io.location import SpecificLocation
from syft.core.node.common.action.garbage_collect_object_action import (
    GarbageCollectObjectAction,
)


def test_garbage_collection_object_action_serde() -> None:
    uid = UID()
    addr = Address(network=SpecificLocation(), device=SpecificLocation())

    msg = GarbageCollectObjectAction(id_at_location=uid, address=addr)

    blob = serialize(msg)

    msg2 = sy.deserialize(blob=blob)

    assert msg2.id_at_location == msg.id_at_location
    assert msg2.address == msg.address



# stdlib
import gc

# third party
import pytest
import torch

# syft absolute
import syft as sy


@pytest.mark.slow
def test_same_var_for_ptr_gc(
    node: sy.VirtualMachine, client: sy.VirtualMachineClient
) -> None:
    """
    Test for checking if the gc is correctly triggered
    when the last reference to the ptr is overwritten
    """
    x = torch.tensor([1, 2, 3, 4])

    for _ in range(100):
        """
        Override the ptr multiple times to make sure we trigger
        the gc
        """
        ptr = x.send(client, pointable=False)

    assert len(node.store) == 1

    ptr.get()
    gc.collect()

    assert len(node.store) == 0


def test_send_same_obj_gc(
    node: sy.VirtualMachine, root_client: sy.VirtualMachineClient
) -> None:
    """
    Test if sending the same object multiple times, register the
    object multiple times - because each send generated another
    remote object
    """
    x = torch.tensor([1, 2, 3, 4])
    ptr = []

    for _ in range(100):
        ptr.append(x.send(root_client, pointable=False))

    gc.collect()
    assert len(node.store) == 100

    del ptr

    gc.collect()
    assert len(node.store) == 0



# syft absolute
from syft.lib.python.bool import Bool
from syft.lib.python.dict import Dict
from syft.lib.python.float import Float
from syft.lib.python.int import Int
from syft.lib.python.list import List
from syft.lib.python.string import String
from syft.lib.python.tuple import Tuple

SyFalse = Bool(False)
SyTrue = Bool(True)

PyFalse = False
PyTrue = True


def test_repr() -> None:
    assert repr(SyFalse) == String("False")
    assert repr(SyTrue) == String("True")
    assert eval(repr(SyFalse)) == SyFalse
    assert eval(repr(SyTrue)) == SyTrue


def test_str() -> None:
    assert str(SyFalse) == String("False")
    assert str(SyTrue) == String("True")


def test_int() -> None:
    assert int(SyFalse) == Int(0)
    assert int(SyFalse) is not SyFalse
    assert int(SyTrue) == 1
    assert int(SyTrue) is not SyTrue


def test_float() -> None:
    assert SyFalse.__float__() == Float(0.0)
    assert SyFalse.__float__() == SyFalse
    assert SyTrue.__float__() == Float(1.0)
    assert SyTrue.__float__() == SyTrue


def test_math() -> None:
    assert +SyFalse == Int(0)
    assert +SyFalse == SyFalse
    assert -SyFalse == Int(0)
    assert -SyFalse == SyFalse
    assert abs(SyFalse) == Int(0)
    assert abs(SyFalse) is not SyFalse
    assert +SyTrue == Int(1)
    assert +SyTrue is not SyTrue
    assert -SyTrue == Int(-1)
    assert abs(SyTrue) == Int(1)
    assert abs(SyTrue) is not SyTrue
    assert ~SyFalse == Int(-1)
    assert ~SyTrue == Int(-2)

    assert SyFalse + Int(2) == Int(2)
    assert SyTrue + Int(2) == Int(3)
    assert Int(2) + SyFalse == Int(2)
    assert Int(2) + SyTrue == Int(3)

    assert SyFalse + SyFalse == Int(0)
    assert SyFalse + SyFalse is not SyFalse
    assert SyFalse + SyTrue == Int(1)
    assert SyFalse + SyTrue is not SyTrue
    assert SyTrue + SyFalse == Int(1)
    assert SyTrue + SyFalse is not SyTrue
    assert SyTrue + SyTrue == Int(2)

    assert SyTrue - SyTrue == Int(0)
    assert SyTrue - SyTrue is not SyFalse
    assert SyFalse - SyFalse == Int(0)
    assert SyFalse - SyFalse is not SyFalse
    assert SyTrue - SyFalse == Int(1)
    assert SyTrue - SyFalse is not SyTrue
    assert SyFalse - SyTrue == Int(-1)

    assert SyTrue * Int(1) == Int(1)
    assert SyFalse * Int(1) == Int(0)
    assert SyFalse * Int(1) is not SyFalse

    assert SyTrue / Int(1) == Int(1)
    assert SyTrue / Int(1) is not SyTrue
    assert SyFalse / Int(1) == Int(0)
    assert SyFalse / Int(1) is not SyFalse

    assert SyTrue % Int(1) == Int(0)
    assert SyTrue % Int(1) is not SyFalse
    assert SyTrue % Int(2) == Int(1)
    assert SyTrue % Int(2) is not SyTrue
    assert SyFalse % Int(1) == Int(0)
    assert SyFalse % Int(1) is not SyFalse

    for b in SyFalse, SyTrue:
        for i in Int(0), Int(1), Int(2):
            assert b ** i == Int(b) ** i
            assert b ** i is not Bool(Int(b) ** i)

    assert (Int(1) == Int(1)) == SyTrue
    assert (Int(1) == Int(0)) == SyFalse
    assert (Int(0) < Int(1)) == SyTrue
    assert (Int(1) < Int(0)) == SyFalse
    assert (Int(0) <= Int(0)) == SyTrue
    assert (Int(1) <= Int(0)) == SyFalse
    assert (Int(1) > Int(0)) == SyTrue
    assert (Int(1) > Int(1)) == SyFalse
    assert (Int(1) >= Int(1)) == SyTrue
    assert (Int(0) >= Int(1)) == SyFalse
    assert (Int(0) != Int(1)) == SyTrue
    assert (Int(0) != Int(0)) == SyFalse

    x = List([Int(1)])
    assert (x == x) == SyTrue
    assert (x == x) != SyFalse

    assert (1 in x) == SyTrue
    assert (0 in x) == SyFalse
    assert (1 not in x) == SyFalse
    assert (0 not in x) == SyTrue

    x = Dict({Int(1): Int(2)})
    assert (x is x) == SyTrue
    assert (x is not x) == SyFalse

    assert (Int(1) in x) == SyTrue
    assert (Int(0) in x) == SyFalse
    assert (Int(1) not in x) == SyFalse
    assert (Int(0) not in x) == SyTrue

    assert not SyTrue == SyFalse
    assert not SyFalse == SyTrue


def test_convert() -> None:
    assert Bool(10) == SyTrue
    assert Bool(1) == SyTrue
    assert Bool(-1) == SyTrue
    assert Bool(0) == SyFalse
    assert Bool("hello") == SyTrue
    assert Bool("") == SyFalse
    assert Bool() == SyFalse


def test_format() -> None:
    assert String("%d") % SyFalse == "0"
    assert String("%d") % SyTrue == "1"
    assert String("%x") % SyFalse == "0"
    assert String("%x") % SyTrue == "1"


def test_hasattr() -> None:
    assert hasattr([], "append") == SyTrue
    assert hasattr([], "wobble") == SyFalse


def test_callable() -> None:
    assert callable(len) == SyTrue
    assert callable(1) == SyFalse


def test_isinstance() -> None:
    assert isinstance(SyTrue, Bool) == SyTrue
    assert isinstance(SyFalse, Bool) == SyTrue
    assert isinstance(SyTrue, int) == SyTrue
    assert isinstance(SyFalse, int) == SyTrue
    assert isinstance(1, Bool) == SyFalse
    assert isinstance(0, Bool) == SyFalse


def test_issubclass() -> None:
    assert issubclass(Bool, int) == SyTrue
    assert issubclass(int, Bool) == SyFalse


def test_contains() -> None:
    assert (Int(1) in {}) == SyFalse
    assert (Int(1) in {Int(1): Int(1)}) == SyTrue


def test_string():
    assert String("xyz").endswith("z") == SyTrue
    assert String("xyz").endswith("x") == SyFalse
    assert String("xyz0123").isalnum() == SyTrue
    assert String("@#$%").isalnum() == SyFalse
    assert String("xyz").isalpha() == SyTrue
    assert String("@#$%").isalpha() == SyFalse
    assert String("0123").isdigit() == SyTrue
    assert String("xyz").isdigit() == SyFalse
    assert String("xyz").islower() == SyTrue
    assert String("XYZ").islower() == SyFalse
    assert String("0123").isdecimal() == SyTrue
    assert String("xyz").isdecimal() == SyFalse
    assert String("0123").isnumeric() == SyTrue
    assert String("xyz").isnumeric() == SyFalse
    assert String(" ").isspace() == SyTrue
    assert String("\xa0").isspace() == SyTrue
    assert String("\u3000").isspace() == SyTrue
    assert String("XYZ").isspace() == SyFalse
    assert String("X").istitle() == SyTrue
    assert String("x").istitle() == SyFalse
    assert String("XYZ").isupper() == SyTrue
    assert String("xyz").isupper() == SyFalse
    assert String("xyz").startswith("x") == SyTrue
    assert String("xyz").startswith("z") == SyFalse


def test_boolean():
    assert SyTrue & Int(1) == Int(1)
    assert not isinstance(SyTrue & Int(1), Bool)
    assert SyTrue & SyTrue == SyTrue

    assert SyTrue | Int(1) == Int(1)
    assert not isinstance(SyTrue | Int(1), Bool)
    assert SyTrue | SyTrue == SyTrue

    assert SyTrue ^ Int(1) == Int(0)
    assert not isinstance(SyTrue ^ Int(1), Bool)
    assert SyTrue ^ SyTrue == SyFalse


def test_types():
    # types are always SyTrue.
    for t in [
        bool,
        complex,
        dict,
        float,
        int,
        list,
        object,
        set,
        str,
        tuple,
        type,
        Int,
        Float,
        Dict,
        List,
        String,
        Tuple,
        Float,
    ]:
        assert bool(t) == SyTrue


def test_operator():
    # stdlib
    import operator

    assert operator.truth(0) == SyFalse
    assert operator.truth(1) == SyTrue
    assert operator.not_(1) == SyFalse
    assert operator.not_(0) == SyTrue
    assert operator.contains([], 1) == SyFalse
    assert operator.contains([1], 1) == SyTrue
    assert operator.lt(0, 0) == SyFalse
    assert operator.lt(0, 1) == SyTrue
    assert operator.is_(SyTrue, SyTrue) == SyTrue
    assert operator.is_(SyTrue, SyFalse) == SyFalse
    assert operator.is_not(SyTrue, SyTrue) == SyFalse
    assert operator.is_not(SyTrue, SyFalse) == SyTrue


def test_from_bytes():
    assert Bool.from_bytes(b"\x00" * 8, "big") == SyFalse
    assert Bool.from_bytes(b"abcd", "little") == SyTrue


def test_sane_len():
    # this test just tests our assumptions about __len__
    # this will start failing if __len__ changes assertions
    for badval in [String("illegal"), Int(-1), Int(1 << 32)]:

        class A:
            def __len__(self):
                return badval

        try:
            Bool(A())
        except (Exception) as e_bool:
            try:
                len(A())
            except (Exception) as e_len:
                assert str(e_bool) == str(e_len)


def test_real_and_imag():
    # TODO add support for proprieties on these
    assert SyTrue.real() == Int(1)
    assert SyTrue.imag() == Int(0)
    assert type(SyTrue.real()) is Int
    assert type(SyTrue.imag()) is Int
    assert SyFalse.real() == Int(0)
    assert SyFalse.imag() == Int(0)
    assert type(SyFalse.real()) is Int
    assert type(SyFalse.imag()) is Int


def test_upcast():
    assert SyFalse.upcast() is False
    assert SyTrue.upcast() is True


def test_abs():
    assert abs(SyTrue) == 1
    assert abs(SyFalse) == 0


def test_add():
    assert SyFalse + SyFalse == 0
    assert SyTrue + SyFalse == 1
    assert SyFalse + SyTrue == 1
    assert SyTrue + SyTrue == 2

    assert SyTrue + 42 == 43
    assert SyFalse + 42 == 42


def test_ceil():
    assert SyTrue.__ceil__() == 1
    assert SyFalse.__ceil__() == 0


def test_divmod():
    r, q = SyTrue.__divmod__(5)
    assert type(r) is Int
    assert type(q) is Int

    assert r == 0
    assert q == 1


def test_floor():
    res = SyTrue.__floor__()
    assert type(res) is Int
    assert res == 1


def test_floordiv():
    res = SyTrue.__floordiv__(5)
    assert type(res) is Int
    assert res == 0


def test_cond():
    assert SyTrue > SyFalse
    assert SyFalse < SyTrue
    assert SyTrue >= SyTrue
    assert SyFalse >= SyFalse
    assert SyTrue <= SyTrue
    assert SyFalse <= SyFalse


def test_hash():
    res = SyTrue.__hash__()
    assert isinstance(res, Int)


def test_invert():
    assert isinstance(SyTrue.__invert__(), Int)
    assert SyTrue.__invert__() == PyTrue.__invert__()
    assert SyFalse.__invert__() == PyFalse.__invert__()


def test_shift():
    assert SyTrue.__lshift__(42) == PyTrue.__lshift__(42)
    assert SyFalse.__rshift__(42) == PyFalse.__rshift__(42)
    assert SyTrue.__rlshift__(42) == PyTrue.__rlshift__(42)
    assert SyFalse.__rrshift__(42) == PyFalse.__rrshift__(42)


def test_mod():
    assert SyTrue.__mod__(42) == PyTrue.__mod__(42)
    assert SyFalse.__mod__(42) == PyFalse.__mod__(42)


def test_mul():
    assert SyTrue.__mul__(42) == PyTrue.__mul__(42)
    assert SyFalse.__mul__(42) == PyFalse.__mul__(42)


def test_ne():
    assert SyTrue != SyFalse
    assert SyFalse != SyTrue


def test_neg():
    assert SyTrue.__neg__() == PyTrue.__neg__()
    assert SyFalse.__neg__() == PyFalse.__neg__()


def test_pos():
    assert SyTrue.__pos__() == PyTrue.__pos__()
    assert SyFalse.__pos__() == PyFalse.__pos__()


def test_protobuf_schema():
    assert Bool.get_protobuf_schema()


def test_to_bytes():
    assert Bool.from_bytes(SyTrue.to_bytes(4, "big"), "big")
    assert not Bool.from_bytes(SyFalse.to_bytes(4, "big"), "big")


def test_hash_bool():
    assert SyTrue.__hash__() != SyFalse.__hash__()


def test_int_bool():
    assert int(SyTrue) == 1
    assert int(SyFalse) == 0


def test_repr_bool():
    assert repr(SyTrue) == repr(True)
    assert repr(SyFalse) == repr(False)


def test_str_bool():
    assert str(SyTrue) == str(True)
    assert str(SyFalse) == str(False)



# syft relative
from ..federated import JSONDict
from ..federated.model_centric_fl_base import ModelCentricFLBase
from ..lib.python import List
from ..lib.torch.module import Module as SyModule
from ..proto.core.plan.plan_pb2 import Plan
from .model_serialization import deserialize_model_params
from .model_serialization import wrap_model_params


class ModelCentricFLClient(ModelCentricFLBase):
    def host_federated_training(
        self,
        model: SyModule,
        client_plans: JSONDict,
        client_protocols: JSONDict,
        client_config: JSONDict,
        server_averaging_plan: Plan,
        server_config: JSONDict,
    ) -> JSONDict:

        # store raw tensors only (not nn.Parameters, no grad)
        # TODO migrate to syft-core protobufs
        params = []
        model_parameters = model.parameters()
        if model_parameters is not None:
            params = [getattr(p, "data", None) for p in model_parameters]
        serialized_model = self.hex_serialize(wrap_model_params(params))

        serialized_plans = self._serialize_dict_values(client_plans)
        serialized_protocols = self._serialize_dict_values(client_protocols)
        serialized_avg_plan = self.hex_serialize(server_averaging_plan)

        # "model-centric/host-training" request body
        message = {
            "type": "model-centric/host-training",
            "data": {
                "model": serialized_model,
                "plans": serialized_plans,
                "protocols": serialized_protocols,
                "averaging_plan": serialized_avg_plan,
                "client_config": client_config,
                "server_config": server_config,
            },
        }

        return self._send_msg(message)

    def retrieve_model(
        self, name: str, version: str, checkpoint: str = "latest"
    ) -> List:
        params = {
            "name": name,
            "version": version,
            "checkpoint": checkpoint,
        }
        serialized_model = self._send_http_req(
            "GET", "/model-centric/retrieve-model", params
        )
        # TODO migrate to syft-core protobufs
        return deserialize_model_params(serialized_model)



# third party
import pytest

# syft absolute
import syft as sy
from syft import serialize


@pytest.mark.asyncio
async def test_client_from_metadata() -> None:

    domain = sy.Domain(name="duet")

    client_metadata = domain.get_metadata_for_client()

    spec_location, name, id = sy.DomainClient.deserialize_client_metadata_from_node(
        metadata=serialize(client_metadata)
    )

    assert domain.domain == spec_location
    assert name == "duet"
    assert id == domain.id



# third party
import pandas as pd

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...lib.python.dict import Dict
from ...lib.python.primitive_factory import PrimitiveFactory
from ...proto.lib.pandas.series_pb2 import PandasSeries as PandasSeries_PB


def object2proto(obj: pd.Series) -> PandasSeries_PB:
    series_dict = PrimitiveFactory.generate_primitive(value=obj.to_dict())
    dict_proto = series_dict._object2proto()

    return PandasSeries_PB(
        id=dict_proto.id,
        series=dict_proto,
    )


def proto2object(proto: PandasSeries_PB) -> pd.Series:
    series_dict = Dict._proto2object(proto=proto.series)
    return pd.Series(series_dict.upcast())


GenerateWrapper(
    wrapped_type=pd.Series,
    import_path="pandas.Series",
    protobuf_scheme=PandasSeries_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)
